"""
Image mask creation and manipulation.
"""
import numpy as np

import cv2

import alphashape as ap

import numba

@numba.jit(nopython=True)
def crossMask(size, ylim=[0,0], xlim=[0,0], channels=3):
    """
    Create a mask in the shape of a cross, in which the inside of the cross is
    the desired region.

    A stripe (horizontal or vertical) can be made by only supplying a single set
    of limits, either for x (vertical) or y (horizontal).

    Parameters
    ----------

    size : [int, int] or [int, int, int]
        The size of the mask to be created (height, width[, channels]). Can include number of channels.
        If not included, default is 3; see `channels`.

    ylim : [int, int]
        The y coordinates of the (beginning and end of the) desired horizontal region

    xlim : [int, int]
        The x coordinates of the (beginning and end of the) desired vertical region

    channels : int
        The number of channels to create for the mask, if not included in `size` parameter.

    Returns
    -------

    np.uint8[3] : Mask array (height, width, channels)
    """

    # If channels are included in the size, use that
    if len(size) == 3:
        channels = size[2]

    maskArr = np.zeros((size[0], size[1], channels), dtype=np.uint8)

    for i in range(size[0]):
        for j in range(size[1]):
            if (i >= ylim[0] and i <= ylim[1]):
                maskArr[i,j] = np.ones(channels)
            elif (j >= xlim[0] and j <= xlim[1]):
                maskArr[i,j] = np.ones(channels)

    return maskArr


@numba.jit(nopython=True)
def rectMask(size, corner=[0,0], dimensions=[0,0], channels=3):
    """
    Create a mask in the shape of a (non-rotated) rectangle, in which the inside
    of the rectangle is the desired region.

    Parameters
    ----------

    size : [int, int] or [int, int, int]
        The size of the mask to be created (height, width[, channels]). Can include
        number of channels. If not included, default is 3; see `channels`.

    corner : [int, int]
        The [y, x] coordinates of the top-left corner of the desired region. 

    dimensions : [int, int]
        The [height, width] of the desired region. If either is set to 0, the
        entire rest of the image will be included in that direction.

    channels : int
        The number of channels to create for the mask, if not included in `size` parameter.

    Returns
    -------

    np.uint8[3] : Mask array (height, width, channels)
    """

    # If channels are included in the size, use that
    # Though if the channels variable is explicitly set to None, that overrides this
    if len(size) == 3:
        channels = size[2]

    xlim = [corner[1], corner[1] + dimensions[1]]
    ylim = [corner[0], corner[0] + dimensions[0]]

    maskArr = np.zeros((size[0], size[1], channels), dtype=np.uint8)
    
    for i in range(size[0]):
        for j in range(size[1]):
            if (i >= ylim[0] and i <= ylim[1] and j >= xlim[0] and j <= xlim[1]):
                maskArr[i,j] = np.ones(channels)

    return maskArr

@numba.jit(nopython=True)
def circularMask(size, center, radius, channels=3):
    """
    Create a mask in the shape of a circle, in which the inside of the circle is
    the desired region.

    Parameters
    ----------

    size : [int, int] or [int, int, int]
        The size of the mask to be created (height, width[, channels]). Can include number of channels.
        If not included, default is 3; see `channels`.

    center : [int, int]
        The [y, x] coordinates of the center of the desired region. 

    radius : int
        The radius of the desired region.

    channels : int
        The number of channels to create for the mask, if not included in `size` parameter.

    Returns
    -------

    np.uint8[3] : Mask array (height, width, channels)
    """

    # If channels are included in the size, use that
    if len(size) == 3:
        channels = size[2]

    # numba doesn't support ogrid, so we just do it manually
    #Y, X = np.ogrid[:size[0], :size[1]]
    Y = np.arange(size[0]).reshape((size[0], 1)) # Column vector
    X = np.arange(size[1]).reshape((1, size[1])) # Row vector

    dist_from_center = np.sqrt((X - center[1])**2 + (Y-center[0])**2)

    maskArr = np.zeros((size[0], size[1], channels), np.uint8)
    for i in range(channels):
        maskArr[:,:,i] = dist_from_center <= radius

    return maskArr


@numba.jit(nopython=True)
def ellipticalMask(size, majorAxisEndpoint1, majorAxisEndpoint2, minorAxisLength, channels=3):
    """
    Create a mask in the shape of a circle, in which the inside of the circle is
    the desired region.

    Parameters
    ----------

    size : [int, int] or [int, int, int]
        The size of the mask to be created (height, width[, channels]). Can include number of channels.
        If not included, default is 3; see `channels`.

    majorAxisEndpoint1 : [int, int]
        The first of two sets of [y, x] coordinates that define the major axis of the desired region. 
        Order does not matter.

    majorAxisEndpoint2 : [int, int]
        The second of two sets of [y, x] coordinates that define the major axis of the desired region. 
        Order does not matter.

    minorAxisLength : int
        The length of the minor axis, or the extent of the shorter dimension of the ellipse.

    channels : int
        The number of channels to create for the mask, if not included in `size` parameter.

    Returns
    -------

    np.uint8[3] : Mask array (height, width, channels)
    """

    minorAxisLength = float(minorAxisLength)

    # If channels are included in the size, use that
    # Though if the channels variable is explicitly set to None, that overrides this
    if len(size) == 3:
        channels = size[2]

    # Center of ellipse
    center = np.array([(majorAxisEndpoint1[0] + majorAxisEndpoint2[0])/2,
              (majorAxisEndpoint1[1] + majorAxisEndpoint2[1])/2], dtype=np.float64)

    # Unit vector along major axis
    eMajor = np.array([majorAxisEndpoint1[0] - majorAxisEndpoint2[0],
                       majorAxisEndpoint1[1] - majorAxisEndpoint2[1]], dtype=np.float64)
    eMajor /= np.sqrt(eMajor[0]**2 + eMajor[1]**2)

    semiMajorAxis = np.sqrt((majorAxisEndpoint1[0] - majorAxisEndpoint2[0])**2 + (majorAxisEndpoint1[1] - majorAxisEndpoint2[1])**2)/2
    semiMinorAxis = minorAxisLength/2

    # Distance from center to either focus part of a triangle with axes
    focalDistance = np.sqrt(semiMajorAxis**2 - semiMinorAxis**2)

    # Foci of the ellipse
    foci = np.zeros((2, 2))
    foci[0,0] = center[0] + focalDistance * eMajor[0]
    foci[0,1] = center[1] + focalDistance * eMajor[1]
    foci[1,0] = center[0] - focalDistance * eMajor[0]
    foci[1,1] = center[1] - focalDistance * eMajor[1]
    #foci = [[center[0] + focalDistance * eMajor, center[1] + focalDistance * eMajor],
    #        [center[0] - focalDistance * eMajor, center[1] - focalDistance * eMajor]]

    # For any point outside of the ellipse, the sum of distance from both foci will
    # be greater than the length of the major axis

    # numba doesn't support ogrid, so we just do it manually
    #Y, X = np.ogrid[:size[0], :size[1]]
    Y = np.arange(size[0]).reshape((size[0], 1)) # Column vector
    X = np.arange(size[1]).reshape((1, size[1])) # Row vector

    fociDistanceSum = np.sqrt((X - foci[0][1])**2 + (Y - foci[0][0])**2) + np.sqrt((X - foci[1][1])**2 + (Y - foci[1][0])**2) 
    
    maskArr = np.zeros((size[0], size[1], channels), dtype=np.uint8)
    for i in range(channels):
        maskArr[:,:,i] = fociDistanceSum <= 2*semiMajorAxis

    return maskArr


@numba.jit(nopython=True)
def mergeMasks(listOfMasks, signs=None, channels=3):
    """
    Merge several masks together into a single new mask, serving as the union of the
    previous set.

    Parameters
    ----------

    listOfMasks : iterable of np.uint8[2] or np.uint8[3]
        A list of the masks to be merged together. If they have different numbers of channels, the
        union will follow the `channels` parameter; otherwise, original shape will be preserved.

    signs : iterable of int
        Array of 1 or -1 that indicates which masks should be added to the final result, and which should
        be subtracted. By default, all are added.

    channels : int
        The number of channels to create for the final mask; used only if not consistent in `listOfMasks` parameter.

    Returns
    -------

    np.uint8[2] or np.uint8[3] : Mask array; can include channels in last dim, or not.
    """

    if signs is None:
        signs = np.ones(len(listOfMasks), dtype=np.int16)

    sumMask = np.zeros((*np.shape(listOfMasks[0])[:2], channels), dtype=np.int16)

    for i in range(len(listOfMasks)):
        for j in range(channels):
            # Assume shape of [H, W, C]
            sumMask[:,:,j] += listOfMasks[i][:,:,0] * signs[i]
                

    #sumMask += np.abs(np.min(sumMask))
    #sumMask = np.ceil(sumMask / np.max(sumMask))
    return (sumMask > 0).astype(np.uint8)


def identifyFeatureBoundary(image,
                            approxFeatureLocation,
                            featureSize,
                            meanThresholdFactor=2,
                            minIslandSize=5,
                            pointDownsampleFactor=5,
                            concaveHullParameter=0,
                            debug=False):
    """
    Detect the boundary of a feature near the specified location.

    Feature detection is done using the same algorithm as in
    `pepe.preprocess.maskFeature()`. Feature should be
    brighter than the background.

    Parameters
    ----------
    image : numpy.ndarray[H,W[,C]]
        Input image to identify feature in. If multi-channel,
        the channels will be averaged (gray-scaled).

    approxFeatureLocation : (float, float)
        Approximate coordinates (y,x) of the location of the feature.

    meanThresholdFactor : float
        Parameter for determining which pixels are a part of the feature.
        Lower value means the detected feature will be larger.

    minIslandSize : int
        Parameter for culling pixels initially detected to be a part of
        the feature, but likely are not actually. eg. If a value of 5,
        is chosen, only pixels surrounded by 24 other pixels detected to
        be a part of the feature will be kept.

        If not an odd number, the value will be rounded up to the next odd
        number.

    pointDownsampleFactor : int >= 1
        Parameter for generating the hull around the points detected to be
        a part of the feature. Higher value means more approximate shape,
        but faster calculations. Value of `1` will give the most accurate
        result.

    concaveHullParamter : float
        Alphashape parameter for generating the hull. A value of `0`
        means the hull will be convex, a value `>= 0` will mean the
        hull is increasingly concave.

        For more information, see [alphashape documentation](https://alphashape.readthedocs.io/en/latest/readme.html)
        related to the 'alpha parameter'.

    debug : bool
        Whether to plot debug information about the feature
        detection once completed or not.


    Returns
    -------
    boundaryPoints : numpy.ndarray[N,2]
        List of N points that comprise the boundary ordered
        such that the boundary can be drawn by drawing all pairs
        of points (i,i+1) (and (-1,0) for the final line).
    """
    # Grayscale if necessary
    if len(np.shape(image)) == 3:
        grayImage = np.mean(image, axis=-1)
    else:
        grayImage = image

    # Crop
    croppedImage = grayImage[approxFeatureLocation[0]-int(featureSize):approxFeatureLocation[0]+int(featureSize),
                                approxFeatureLocation[1]-int(featureSize):approxFeatureLocation[1]+int(featureSize)]

    # Identify regions that are above the mean
    thresholdedImage = np.float64(croppedImage > meanThresholdFactor*np.mean(croppedImage))

    # Blur and then threshold again, meaning that anything who isn't surrounded by
    # `minIslandSize/2` neighbors in each direction gets removed
    kernel = minIslandSize + int(not minIslandSize%2) # Make sure it is odd
    thresholdedImage = np.float64(cv2.blur(thresholdedImage, (kernel,kernel)) == 1)

    # Generate a convex hull
    alphashape = ap.alphashape(np.array(np.where(thresholdedImage)).T[::pointDownsampleFactor], concaveHullParameter)
    
    boundaryPoints = np.array(alphashape.boundary.coords)[:,:2:][:,::-1]

    if debug:
        fig, ax = plt.subplots(1, 2, figsize=(8,4))

        ax[0].imshow(thresholdedImage)
        ax[1].imshow(grayImage)
        for i in range(len(boundaryPoints-1)):
            ax[1].plot(boundaryPoints[i:i+2,0], boundaryPoints[i:i+2,1], c='red', alpha=1)

        ax[0].set_title('Thresholded Image')
        ax[1].set_title('Identified Feature Boundary')
        plt.show()

    # Convert local coordinates back to absolute image coordinates
    return approxFeatureLocation + boundaryPoints - featureSize

def maskFeature(image,
                approxFeatureLocation,
                featureSize,
                meanThresholdFactor=1,
                minIslandSize=5,
                pointDownsampleFactor=5,
                concaveHullParameter=0,
                extraPadding=0,
                allowRecurse=True,
                debug=False):
    """
    Generate a mask to remove a detected feature near the specified location.

    Feature detection is done using the same algorithm as in
    `pepe.preprocess.detectFeatureBoundary()`. Feature should be
    brighter than the background.

    Parameters
    ----------
    image : numpy.ndarray[H,W[,C]]
        Input image to identify feature in. If multi-channel,
        the channels will be averaged (gray-scaled).

    approxFeatureLocation : (float, float)
        Approximate coordinates (y,x) of the location of the feature.

    meanThresholdFactor : float
        Parameter for determining which pixels are a part of the feature.
        Lower value means the detected feature will be larger.

    minIslandSize : int
        Parameter for culling pixels initially detected to be a part of
        the feature, but likely are not actually. eg. If a value of 5,
        is chosen, only pixels surrounded by 24 other pixels detected to
        be a part of the feature will be kept.

        If not an odd number, the value will be rounded up to the next odd
        number.

    pointDownsampleFactor : int >= 1
        Parameter for generating the hull around the points detected to be
        a part of the feature. Higher value means more approximate shape,
        but faster calculations. Value of `1` will give the most accurate
        result.

    concaveHullParamter : float
        Alphashape parameter for generating the hull. A value of `0`
        means the hull will be convex, a value `>= 0` will mean the
        hull is increasingly concave.

        For more information, see [alphashape documentation](https://alphashape.readthedocs.io/en/latest/readme.html)
        related to the 'alpha parameter'.

    extraPadding : int
        Extra padding at the end of the feature identification
        added to the shape.

    allowRecurse : bool
        If the  detected feature is larger than the size,
        and this value is `True`, the function will recursively
        call itself (at most once) with a larger feature size to try
        and detect the entire feature.
        
    debug : bool
        Whether to plot debug information about the feature
        detection once completed or not.


    Returns
    -------
    imageMask : numpy.ndarray[H,W]
        Image mask with values of `1` everywhere
        except for points that are detected to be a
        part of the feature, which will have values
        of `0`.

    """
    # Grayscale if necessary
    if len(np.shape(image)) == 3:
        grayImage = np.mean(image, axis=-1)
    else:
        grayImage = image

    # Crop
    croppedImage = grayImage[max(approxFeatureLocation[0]-int(featureSize), 0):min(approxFeatureLocation[0]+int(featureSize), grayImage.shape[0]),
                                max(approxFeatureLocation[1]-int(featureSize), 0):min(approxFeatureLocation[1]+int(featureSize), grayImage.shape[1])]

    # Identify regions that are above the mean
    thresholdedImage = np.float64(croppedImage > meanThresholdFactor*np.mean(croppedImage))

    # Blur and then threshold again, meaning that anything who isn't surrounded by
    # `minIslandSize/2` neighbors in each direction gets removed
    kernel = minIslandSize + int(not minIslandSize%2) # Make sure it is odd
    thresholdedImage = np.float64(cv2.blur(thresholdedImage, (kernel,kernel)) == 1)

    # Generate a convex hull
    alphashape = ap.alphashape(np.array(np.where(thresholdedImage)).T[::pointDownsampleFactor], concaveHullParameter)

    if alphashape is not None:
        boundaryPoints = np.array(alphashape.boundary.coords)[:,:2:][:,::-1]
    else:
        boundaryPoints = None

    # If any of the boundary points are at the edge of the image, that means the 
    # feature extends beyond the cropped region, and we should try again with a slightly
    # larger region
    edgePadding = 5
    if (True in (boundaryPoints.flatten() < edgePadding) or True in (boundaryPoints.flatten() + edgePadding > 2*featureSize)) and allowRecurse:
        # We could try and recenter the crop window, but that doesn't actually work all that well...
        # Maybe since averaging the boundary points doesn't actually calculate the center of mass
        #newFeatureCenter = approxFeatureLocation + np.int64(np.mean(boundaryPoints, axis=0)) - featureSize
        newFeatureCenter = approxFeatureLocation
        return maskFeature(image, newFeatureCenter, featureSize*1.5,
                           meanThresholdFactor, minIslandSize, pointDownsampleFactor,
                           concaveHullParameter, False, debug)
        
    maskImage = np.zeros_like(croppedImage, dtype=np.uint8)

    for i in range(len(boundaryPoints)-1):
        cv2.line(maskImage, np.int64(boundaryPoints[i]), np.int64(boundaryPoints[i+1]), 1, minIslandSize)        
    cv2.line(maskImage, np.int64(boundaryPoints[-1]), np.int64(boundaryPoints[0]), 1, minIslandSize)
                    
    outsidePoint = (0,0)
    cv2.floodFill(maskImage, None, outsidePoint, 2)

    # Add extra padding
    if extraPadding > 0:
        oddPadding = extraPadding + int(not extraPadding%2) # Make sure it is odd
        maskImage = cv2.blur(maskImage, (oddPadding, oddPadding))

    maskImage = maskImage == 2

    fullImageMask = np.ones(image.shape[:2], dtype=np.uint8)
    fullImageMask[max(approxFeatureLocation[0]-int(featureSize), 0):min(approxFeatureLocation[0]+int(featureSize), grayImage.shape[0]),
                                max(approxFeatureLocation[1]-int(featureSize), 0):min(approxFeatureLocation[1]+int(featureSize), grayImage.shape[1])] = maskImage

    if debug:
        fig, ax = plt.subplots(1, 3, figsize=(12,4))

        ax[0].imshow(thresholdedImage)
        ax[1].imshow(croppedImage)
        for i in range(len(boundaryPoints)-1):
            ax[1].plot(boundaryPoints[i:i+2,0], boundaryPoints[i:i+2,1], c='red', alpha=1)

        ax[2].imshow(fullImageMask)
        
        ax[0].set_title('Thresholded Image')
        ax[1].set_title('Identified Feature Boundary')
        ax[2].set_title('Full Image Mask')
        plt.show()

    return fullImageMask


def polygonMask(boundaryPoints, maskSize, debug=False):
    """
    Create a mask of a polygon defined by the points that
    comprise its boundary.

    Parameters
    ----------
    boundaryPoints : iterable of points
        Iterable of points (y,x) that define the
        boundary of the polygon. Should be ordered
        such that drawing each pair of points (i,i+1)
        creates the desired shape.

    maskSize : [H,W]
        The size of the mask array to create.

    debug : bool
        Whether or not to plot a visualization of
        the mask once it is completed.

    Returns
    -------
    maskImage : numpy.ndarray[H,W]
        Mask array, with values of `1` inside
        the polygon, and `0` outside.

    """
    maskImage = np.ones(maskSize, dtype=np.uint8)

    for i in range(len(boundaryPoints)-1):
        # Have to reverse coordinates
        cv2.line(maskImage, np.int64(boundaryPoints[i][::-1]), np.int64(boundaryPoints[i+1][::-1]), 255, 1)        
    cv2.line(maskImage, np.int64(boundaryPoints[-1][::-1]), np.int64(boundaryPoints[0][::-1]), 255, 1)
                    
    outsidePoint = (0,0)
    cv2.floodFill(maskImage, None, outsidePoint, 255)

    maskImage = (maskImage != 255).astype(np.float64)

    if debug:
        plt.imshow(maskImage)
        plt.show()

    return maskImage


def removePadding(image):
    """
    Remove padding from an image.

    Parameters
    ----------
    image : array-like
        Image to remove excess padding from.

    Returns
    -------
    unpaddedImage : array-like
        Original image with rows and columns
        removed if they contained no image data.
    """
    # Grayscale if necessary
    if len(np.shape(image)) == 3:
        grayImage = np.mean(image, axis=-1)
    else:
        grayImage = image

    rowSum = np.sum(grayImage, axis=0)
    columnSum = np.sum(grayImage, axis=-1)

    # There is probably a much fancier way to do this, but this
    # is fast enough, so... essentially we just count until the sum
    # is not zero anymore (ie there is some image data)
    
    leftEdge = None
    for i in range(len(rowSum)):
        leftEdge = i
        if rowSum[i] > 0:
            break

    rightEdge = None
    for i in range(len(rowSum))[::-1]:
        rightEdge = i
        if rowSum[i] > 0:
            break

    topEdge = None
    for i in range(len(columnSum)):
        topEdge = i
        if columnSum[i] > 0:
            break
            
    bottomEdge = None
    for i in range(len(columnSum))[::-1]:
        bottomEdge = i
        if columnSum[i] > 0:
            break

    return image[topEdge:bottomEdge,leftEdge:rightEdge]




"""
Basic methods for computing metrics based on the pixel intensity within images.
"""

import numpy as np

import cv2
import numba

from pepe.preprocess.Image import checkImageType

def averageBrightness(frame):
    r"""
    Compute the average brightness (pixel intensity) of a frame.

    Parameters
    ----------

    frame : np.ndarray[H,W] or str
        The frame for which to compute the average brightness. Can also
        be a path to an image.


    Returns
    -------

    averageBrightness : float
        The average brightness of the image.
    """
    return np.mean(checkImageType(frame))


def varianceBrightness(frame):
    r"""
    Compute the standard deviation in brightness (pixel intensity) of a frame.

    Parameters
    ----------

    frame : np.ndarray[H,W] or str
        The frame for which to compute the average brightness. Can also
        be a path to an image.


    Returns
    -------

    varianceBrightness: float
        The standard deviation of the brightness of the image.
    """
    return np.var(checkImageType(frame))


@numba.jit(nopython=True)
def gSquared(properFrame):
    """
    The gradient squared at each pixel of the image, also known as the convolution
    of a Laplacian kernel across the image.

    Optimized via `numba`.

    Edge values are padded with values of 0.

    Parameters
    ----------

    properFrame : np.ndarray[H,W]
        An array representing a single channel of an image.


    Returns
    -------

    gSqr : np.ndarray[H,W]
        The gradient squared at every point.


    References
    ----------

    [1] DanielsLab Matlab implementation, https://github.com/DanielsNonlinearLab/Gsquared

    [2] Abed Zadeh, A., Bares, J., Brzinski, T. A., Daniels, K. E., Dijksman, J., Docquier, N.,
    Everitt, H. O., Kollmer, J. E., Lantsoght, O., Wang, D., Workamp, M., Zhao, Y., & Zheng, H.
    (2019). Enlightening force chains: A review of photoelasticimetry in granular matter. Granular
    Matter, 21(4), 83. [10.1007/s10035-019-0942-2](https://doi.org/10.1007/s10035-019-0942-2)
    """
    
    # Take the full size of the image, though know that the outermost row and
    # column of pixels will be 0
    gSquared = np.zeros_like(properFrame)

    # Iterate over every pixel
    # Regardless of whether we have a good region, G^2 needs a buffer of 1 pixel on each
    # side, so we have to crop down more
    for j in range(1, np.shape(properFrame)[0]-1):
        for k in range(1, np.shape(properFrame)[1]-1):
            # I've put a little picture of which pixels we are comparing
            # for each calculation (O is the current pixel, X are the
            # ones we are calculating)

            # - - -
            # X O X
            # - - -
            g1 = float(properFrame[j, k-1]) - float(properFrame[j, k+1])

            # - X -
            # - O -
            # - X -
            g2 = float(properFrame[j-1, k]) - float(properFrame[j+1, k])

            # - - X
            # - O -
            # X - -
            g3 = float(properFrame[j-1, k+1]) - float(properFrame[j+1, k-1])

            # X - -
            # - O -
            # - - X
            g4 = float(properFrame[j-1, k-1]) - float(properFrame[j+1, k+1])

            gSquared[j,k] = (g1*g1/4.0 + g2*g2/4.0 + g3*g3/8.0 + g4*g4/8.0)/4.0

    return gSquared


def averageGSquared(frame):
    r"""
    Compute the average local gradient squared, or \(G^2\), of a frame.

    If multichannel image is provided, will convert to grayscale by averaging
    over the channels.

    Parameters
    ----------

    frame : np.ndarray[H,W] or str
        The frame for which to compute the average gradient squared. Can also
        be a path to an image.


    Returns
    -------

    averageGSquared : float
        The average gradient squared of the image.


    References
    ----------

    [1] DanielsLab Matlab implementation, https://github.com/DanielsNonlinearLab/Gsquared

    [2] Abed Zadeh, A., Bares, J., Brzinski, T. A., Daniels, K. E., Dijksman, J., Docquier, N.,
    Everitt, H. O., Kollmer, J. E., Lantsoght, O., Wang, D., Workamp, M., Zhao, Y., & Zheng, H.
    (2019). Enlightening force chains: A review of photoelasticimetry in granular matter. Granular
    Matter, 21(4), 83. [10.1007/s10035-019-0942-2](https://doi.org/10.1007/s10035-019-0942-2)
    """
    
    # This will load in the image if the method is passed an image
    # file path
    properFrame = checkImageType(frame)

    # Make sure that our image is grayscale
    if properFrame.ndim == 3:
        properFrame = np.mean(properFrame, axis=-1)

    # Use the optimzed gSquared method so it is fast
    return np.mean(gSquared(properFrame))


def varianceGSquared(frame):
    r"""
    Compute the variance in the local gradient squared, or \(G^2\), of a frame.

    If multichannel image is provided, will convert to grayscale by averaging
    over the channels.

    Parameters
    ----------

    frame : np.ndarray[H,W] or str
        The frame for which to compute the variance of the gradient squared. Can also
        be a path to an image.


    Returns
    -------

    varianceGSquared : float
        The variance of the gradient squared of the image.
 

    References
    ----------

    [1] DanielsLab Matlab implementation, https://github.com/DanielsNonlinearLab/Gsquared

    [2] Abed Zadeh, A., Bares, J., Brzinski, T. A., Daniels, K. E., Dijksman, J., Docquier, N.,
    Everitt, H. O., Kollmer, J. E., Lantsoght, O., Wang, D., Workamp, M., Zhao, Y., & Zheng, H.
    (2019). Enlightening force chains: A review of photoelasticimetry in granular matter. Granular
    Matter, 21(4), 83. [10.1007/s10035-019-0942-2](https://doi.org/10.1007/s10035-019-0942-2)
    """

    # This will load in the image if the method is passed an image
    # file path
    properFrame = checkImageType(frame)

    # Make sure that our image is grayscale
    if properFrame.ndim == 3:
        properFrame = np.mean(properFrame, axis=-1)

    # Use the optimzed gSquared method so it is fast
    return np.var(gSquared(properFrame))



"""
Wizard to help user select parameters for various operations. WIP.
"""
import numpy as np
import sys

from PyQt5 import QtWidgets, QtCore, Qt, QtGui

class HelpWindow(QtWidgets.QMainWindow):

    def __init__(self, app):
        super(HelpWindow, self).__init__()
        
        #######################
        #    Basic setup
        #######################
        screen = app.primaryScreen()
        defaultWindowSize = screen.size() * .5
        self.resize(defaultWindowSize)
        #self.setWindowTitle('pepe Parameter Help')
        self.setWindowTitle('float')

        self.cWidget = QtWidgets.QWidget()
        self.vLayout = QtWidgets.QVBoxLayout()
        self.cWidget.setLayout(self.vLayout)

        self.titleLbl = QtWidgets.QLabel()
        self.titleLbl.setText('Howdy!\nWelcome to the pepe parameter selection wizard!')
        self.titleLbl.setFont(QtGui.QFont('Sans', 20))

        self.titleLbl.setAlignment(Qt.Qt.AlignCenter)
        self.vLayout.addWidget(self.titleLbl)

        self.setCentralWidget(self.cWidget)

        self.nextBtn = QtWidgets.QPushButton('Next')
        self.nextBtn.setMaximumWidth(75)
        self.cancelBtn = QtWidgets.QPushButton('Cancel')
        self.cancelBtn.setMaximumWidth(75)
        self.restartBtn = QtWidgets.QPushButton('Restart')
        self.restartBtn.setMaximumWidth(75)

        self.bottomHLayout = QtWidgets.QHBoxLayout()

        self.bottomHLayout.addWidget(self.cancelBtn)
        self.bottomHLayout.addWidget(self.restartBtn)
        self.bottomHLayout.addWidget(self.nextBtn)

        self.vLayout.addLayout(self.bottomHLayout)


        #######################
        #   Parameter Selection
        #######################
        # Here we decide which features we want to select parameters
        # for
        bundledParameters = {'Circle tracking': 'ctrack',
                             'Rotation tracking': 'rtrack',
                             'Masking': 'mask',
                             'Optimization': 'opt'}

        availableParameters = {}


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)

    window = HelpWindow(app)
    window.show()

    sys.exit(app.exec())



"""
Methods to visualize forces acting on a particle.
"""
import numpy as np
import matplotlib.pyplot as plt

from pepe.visualize import visCircles, genRandomDistancedColors, genColors


def visForces(forceArr, alphaArr, betaArr, centerArr=None, angleArr=None, fps=None):
    """
    Visualize all of the forces acting on a single particle, by plotting
    the magnitudes, alphas, and betas (and optionally center position).

    Requires that `forceArr`, `alphaArr`, `betaArr`, and `centerArr` have
    indices of `forceArr[forceIndex][timeIndex]`, as would be given by
    rectangularizing via `pepe.utils.rectangularizeForceArrays()` and 
    indexing a single value in the first dimension.

    Examples
    --------

    ```
    # Solve for some forces
    forceArr, alphaArr, betaArr, centerArr, radiusArr = forceOptimize(...)

    # Particle index
    pI = 0

    fig, ax = visForces(forceArr[pI], alphaArr[pI], betaArr[pI], centerArr[pI])
    plt.show()
    ```

    Parameters
    ----------

    forceArr : np.ndarray[F,T]
        Array containing magnitudes of F forces for T timesteps.

    alphaArr : np.ndarray[F,T]
        Array containing alpha angles of F forces for T timesteps.

    betaArr : np.ndarray[F,T]
        Array containing beta angles of F forces for T timesteps.

    centerArr : np.ndarray[T,2] or None
        Array containing center position in form [y,x] of the particle for T timesteps.

    angleArr : np.ndarray[T] or None
        Array containing angles in radians of the particle for T timesteps.

    fps : float or None
        The number of frames per second of the capture video, used to convert the x-axis
        units from frame number to proper seconds.

    Returns
    -------

    fig : plt.figure()
        The figure object the quantities are plotted on.

    ax : plt.axis()
        The list of 3 (or 4, if centers are provided) axes that the quantities are plotted on.
    """

    fig, ax = plt.subplots(1, 3 + int(centerArr is not None) + int(angleArr is not None), figsize=(3.6*(3+int(centerArr is not None)+int(angleArr is not None)),4))
    
    if len(forceArr) == 0:
        return fig, ax

    tArr = np.arange(len(forceArr[0]))
   
    if fps is None:
        fps = 1
   
    if centerArr is not None:
        ax[-1 - int(angleArr is not None)].plot(tArr/fps, centerArr[:,1], label='X')
        ax[-1 - int(angleArr is not None)].plot(tArr/fps, centerArr[:,0], label='Y')
        ax[-1 - int(angleArr is not None)].set_ylabel('Position [px]')
        ax[-1 - int(angleArr is not None)].legend()

    if angleArr is not None:
        # The weird ax indexing is to make sure it works regardless of
        # whether centerArr is passed or not
        ax[-1].plot(tArr/fps, angleArr)
        ax[-1].set_ylabel('Angle [rad]')

    for i in range(len(forceArr)):
        ax[0].plot(tArr/fps, forceArr[i])
        ax[1].plot(tArr/fps, alphaArr[i], 'o')
        ax[2].plot(tArr/fps, betaArr[i], 'o')

    ax[0].set_ylabel('Force [N]')
    ax[1].set_ylabel('Alpha [rad]')
    ax[2].set_ylabel('Beta [rad]')

    for i in range(3 + int(centerArr is not None) + int(angleArr is not None)):
        if fps == 1:
            ax[i].set_xlabel('Time [frame]')
        else:
            ax[i].set_xlabel('Time [s]')

    fig.tight_layout()

    return fig, ax 


def visContacts(center, radius, betaArr, alphaArr=None, forceArr=None, ax=None, setBounds=False, circleColor=None, forceColors=None, drawCircle=False):
    """
    Visualize the contacts for a system of particles, indicating either positions
    of contacts, or positions and contact angles (if `alphaArr` is provided).

    Returns
    -------

    ax : plt.axis()
    """

    npCenter = np.array(center)

    if ax is None:
        fig, ax = plt.subplots()

    if forceColors is None:
        colors = genColors(len(betaArr), 1000)
    elif type(forceColors) is str:
        colors = [forceColors for i in range(len(betaArr))]
    else:
        colors = forceColors
    
    if drawCircle:
        visCircles([npCenter], [radius], ax, setBounds=setBounds, colors=colors[0])

    if len(betaArr) == 0:
        return ax

    if forceArr is None:
        weighting = np.zeros(len(betaArr)) + radius/4
    else:
        weighting = np.array(forceArr) * radius/4

    if alphaArr is None:
        for i in range(len(betaArr)):
            contactPoint = npCenter + .95*radius * np.array([np.cos(betaArr[i]), np.sin(betaArr[i])])
            ax.scatter([contactPoint[1]], [contactPoint[0]], c=colors[i])
            #c = plt.Circle(contactPoint[::-1], weighting[i], color='red', fill=False, linewidth=1)
            #ax.add_artist(c)
    else:
        for i in range(len(betaArr)):
            if not np.isnan(alphaArr[i]):
                contactPoint = npCenter + radius * np.array([np.cos(betaArr[i]), np.sin(betaArr[i])])
                point1 = contactPoint + weighting[i] * np.array([np.cos(betaArr[i] + alphaArr[i]), np.sin(betaArr[i] + alphaArr[i])])
                point2 = contactPoint - weighting[i] * np.array([np.cos(betaArr[i] + alphaArr[i]), np.sin(betaArr[i] + alphaArr[i])])
                ax.plot([point1[1], point2[1]], [point1[0], point2[0]], linewidth=5, c=colors[i])

    return ax


def fullVisContacts(outputDir, centerArr, radiusArr, betaArr, alphaArr=None, forceArr=None, forceColors=None, circleColors=None, startIndex=0, imageSize=(1024, 1280), fps=25):

    if forceColors is None:
        forceColorArr = [genRandomColors(len(b), int(time.perf_counter()*1e6) % 1024) for b in betaArr]
    else:
        forceColorArr = forceColors

    if len(betaArr) == 0:
        return False

    # The list comprehension is to make sure that we index a particle that actually has forces acting
    # on it.
    tSteps = len([b for b in betaArr if len(b) > 0][0])
    # To save the image of each plot so we can create a gif at the end
    images = [None for i in range(tSteps)]

    for i in range(tSteps):
        clear_output(wait=True)
        fig, ax = plt.subplots()
        
        visCircles([centerArr[p][i] for p in range(len(centerArr))], [radiusArr[p][i] for p in range(len(radiusArr))], ax=ax)

        for particleIndex in range(len(betaArr)):
            visContacts(centerArr[particleIndex][i], radiusArr[particleIndex][i],
                        betaArr[particleIndex][:,i], ax=ax, forceColors=forceColors[particleIndex])#, alphaArr=alphaArr[particleIndex][:,i])
            
        ax.set_xlim([0, imageSize[1]])
        ax.set_ylim([0, imageSize[0]])
        ax.set_aspect('equal')
        ax.set_title(f'Frame {startIndex + i}')
        
        canvas = plt.get_current_fig_manager().canvas
        canvas.draw()
        images[i] = Image.frombytes('RGB', canvas.get_width_height(), 
                     canvas.tostring_rgb())
        plt.close(fig)

    images[0].save(outputDir + 'contact_tracking.gif', save_all=True, append_images=images[1:], duration=fps, optimize=True, loop=True)



"""
Methods to generate good color schemes for visualization.
"""
import numpy as np
import matplotlib.colors as mcolors

from sklearn.neighbors import KDTree

# Some manually selected colors that are good
# Taken from xkcd survey: https://blog.xkcd.com/2010/05/03/color-survey-results/
MANUAL_COLORS = ['#7E1E9C', # Purple
                '#15B01A', # Green
                '#0343DF', # Blue
                '#FF88C0', # Pink
                '#653700', # Brown
                '#E50000', # Red
                '#029386', # Teal
                '#F97306', # Orange
                '#033500', # Dark green
                '#00035B', # Dark blue
                '#650021', # Maroon
                '#BF77F6', # Light purple
                '#929591', # Gray
                '#6E750E', # Olive
                '#00FFFF', # Cyan
                ]


def rgbToHex(rgbC):
    return '%02x%02x%02x' % rgbC

def hexToRgb(hexC):
    return [int(hexC[i:i+2], 16) for i in (0, 2, 4)]

def genColors(size, seed=21):
    """
    Generate a set of colors for visualization by a hierarchy of methods:

    1. First, select from MANUAL_COLORS defined above.
    2. Use genRandomDistancedColors()
    3. Use genRandomColors() 
    """
    if size < len(MANUAL_COLORS):
        return [MANUAL_COLORS[i] for i in range(size)]
    
    else:
        # genRandomDistanceColors will already supplement with full random
        # colors if size is too big, so no need to deal with it here.
        return MANUAL_COLORS + genRandomDistancedColors(size - len(MANUAL_COLORS), seed)


def genRandomColors(size, seed=21):
    """
    Generate colors by randomly selecting rgb values.

    Parameters
    ----------

    size : int
        The number of colors to generate.

    seed : int
        The seed to use for random generation.


    Returns
    -------
    
    colors : list[size]
        A list of hex codes for colors.
    """
    np.random.seed(seed)

    randomColors = [f"#{rgbToHex(tuple(np.random.choice(range(256), size=3).flatten()))}" for i in range(size)]

    return randomColors

def genRandomDistancedColors(size, seed=21):
    """
    Generate colors by sampling from the list of CSS named colors,
    where the first color will be randomly selected, and each subsequent
    one will be chosen to maximize it's distance from previously selected
    colors in LAB color space (using a kd-tree).

    If the number of requested colors is greater than the number of named
    colors, the remaining ones will be randomly generated.

    Maybe a bit overkill...

    Parameters
    ----------

    size : int
        The number of colors to generate.

    seed : int
        The seed to use for random generation.


    Returns
    -------
    
    colors : list[size]
        A list of hex codes for colors.
    """
    allCSSColors = list(mcolors.CSS4_COLORS.values())

    if size > len(allCSSColors):
        allCSSColors += genRandomColors(size - len(allCSSColors), seed)
        return allCSSColors

    np.random.seed(seed)

    allCSSColorsLAB = np.array([rgbToLab(hexToRgb(c[1:])) for c in allCSSColors])

    # Randomly generate the starting one
    selection = np.random.randint(len(allCSSColors))
    addedIndices = [selection]
    colors = [allCSSColors[selection]]

    colorsLAB = np.zeros((size, 3))    
    colorsLAB[0] = allCSSColorsLAB[selection]

    kdTree = KDTree(allCSSColorsLAB, leaf_size=10)
    # For each other color, we want to select the furthest possible color from our current set
    for i in range(1, size):

        dist, ind = kdTree.query(colorsLAB[:i], k=len(allCSSColorsLAB))
        dist = np.array([dist[j,:][np.argsort(ind[j])] for j in range(len(dist))])
        fullCost = np.sum(dist, axis=0)
        selection = np.argsort(fullCost)[::-1]
        selection = [s for s in selection if not s in addedIndices]

        colors.append(allCSSColors[selection[0]])
        addedIndices.append(selection[0])
        colorsLAB[i] = allCSSColorsLAB[selection[0]]

    return colors


def maxDifferenceOrderColors(colors):
    """
    Sort a set of colors such that the difference between consecutive
    colors is maximized.

    A fun little application of the (inverse of the) travelling salesman problem :)
    """
    pass


# This method was taken from:
# https://gist.github.com/manojpandey/f5ece715132c572c80421febebaf66ae
# TODO: Clean it up
def rgbToLab(inputColor):

    num = 0
    RGB = [0, 0, 0]

    for value in inputColor:
        value = float(value) / 255

        if value > 0.04045:
            value = ((value + 0.055) / 1.055) ** 2.4
        else:
            value = value / 12.92

        RGB[num] = value * 100
        num = num + 1

    XYZ = [0, 0, 0, ]

    X = RGB[0] * 0.4124 + RGB[1] * 0.3576 + RGB[2] * 0.1805
    Y = RGB[0] * 0.2126 + RGB[1] * 0.7152 + RGB[2] * 0.0722
    Z = RGB[0] * 0.0193 + RGB[1] * 0.1192 + RGB[2] * 0.9505
    XYZ[0] = round(X, 4)
    XYZ[1] = round(Y, 4)
    XYZ[2] = round(Z, 4)

    # Observer= 2°, Illuminant= D65
    XYZ[0] = float(XYZ[0]) / 95.047         # ref_X =  95.047
    XYZ[1] = float(XYZ[1]) / 100.0          # ref_Y = 100.000
    XYZ[2] = float(XYZ[2]) / 108.883        # ref_Z = 108.883

    num = 0
    for value in XYZ:

        if value > 0.008856:
            value = value ** (0.3333333333333333)
        else:
            value = (7.787 * value) + (16 / 116)

        XYZ[num] = value
        num = num + 1

    Lab = [0, 0, 0]

    L = (116 * XYZ[1]) - 16
    a = 500 * (XYZ[0] - XYZ[1])
    b = 200 * (XYZ[1] - XYZ[2])

    Lab[0] = round(L, 4)
    Lab[1] = round(a, 4)
    Lab[2] = round(b, 4)

    return Lab



import numpy as np

from pepe.preprocess import crossMask
from pepe.simulate import genSyntheticResponse
from pepe.analysis import initialForceSolve, forceOptimize
from pepe.utils import preserveOrderSort

def test_synthetic_force_solve_brightfield():
    """
    """
    # TODO: pass for now, since it isn't consistent
    # enough (inherent issue in non-linear optimization...)
    return

    fSigma = 100.
    pxPerMeter = 10000.
    brightfield = False

    # Parameters of our force solving method
    contactPadding = 10
    g2MaskPadding = 1
    contactMaskRadius = 50

    imageShape = (1024, 1280, 3)

    # Options for optimization
    optimizationKwargs = {"maxEvals": [100, 100, 100], "method": 'nelder',
                      "parametersToFit": [['f'], ['f', 'a'], ['a']],
                      "allowRemoveForces": True, "useTolerance": False,
                      "allowAddForces": True, "minForceThreshold": .001,
                      "localizeAlphaOptimization": False, "imageScaleFactor": .5,
                      "forceBalanceWeighting": .05}

    # For unit testing
    forceErrorTolerance = .1 # [N]
    betaErrorTolerance = .05 # [rad]
    alphaErrorTolerance = .5 # [rad], Much larger, since this one is usually a mess

    # Create a cross-shaped channel
    maskArr = crossMask(imageShape, xlim=np.array([460, 810]), ylim=np.array([350, 700]))

    # Our particles
    centers = np.array([[524, 257], [524, 605], [524, 955], # Center particles
                       [175, 635], [870, 635]], dtype=np.float64) # Top and bottom particles
    radii = np.array([175, 174, 173, 171, 172]) # Odd numbers are good

    # Same order as the centers above
    # Made by just messing around with numbers until things look good
    forces = np.array([[.08, .1, .4], [.4, .05, .32, .39], [.06, .05, .05], [.37, .15], [.35, .16]], dtype='object')
    betas = np.array([[0., np.pi, np.pi/2], [.1, np.pi/2, -np.pi/2, np.pi-.1], [-np.pi/2, 0., np.pi], [-.1, np.pi/2], [-np.pi+.1, np.pi/2]], dtype='object')
    alphas = np.array([[0., 0., 0.], [-.6, 0., 0., .6], [0., 0., 0.], [0., .6], [-.6, 0.]], dtype='object')

    photoelasticChannel = np.zeros(imageShape[:2])

    for i in range(len(centers)):
        photoelasticChannel += genSyntheticResponse(np.array(forces[i]), np.array(alphas[i]), np.array(betas[i]), fSigma, radii[i], pxPerMeter, brightfield, imageSize=imageShape[:2], center=centers[i])

    # Optional: add some noise to the photoelastic channel
    #photoelasticChannel += np.random.uniform(-.2, .2, size=photoelasticChannel.shape)

    photoelasticChannel = np.abs(photoelasticChannel)

    # Put the photoelastic response together with circles
    compositeImage = np.zeros(imageShape, dtype=np.uint8)
    compositeImage[:,:,0] = maskArr[:,:,0]*100
    compositeImage[:,:,1] = np.uint8(photoelasticChannel*255)
    compositeImage[:,:,2] = maskArr[:,:,0]*100

    # Generate initial guess
    forceGuessArr, alphaGuessArr, betaGuessArr = initialForceSolve(photoelasticChannel,
                                                    centers, radii, fSigma, pxPerMeter,
                                                    contactPadding, g2MaskPadding,
                                                    contactMaskRadius=contactMaskRadius,
                                                    boundaryMask=maskArr, ignoreBoundary=False)

    # Optimize forces
    for i in range(len(centers)):
        optForceArr, optBetaArr, optAlphaArr, res = forceOptimize(np.zeros(len(forceGuessArr[i]))+.1, betaGuessArr[i], alphaGuessArr[i],
                                                                  radii[i], centers[i], photoelasticChannel, fSigma, pxPerMeter, brightfield,
                                                                  **optimizationKwargs, debug=False)

        assert len(optForceArr) == len(forces[i]), f'Incorrect number of forces found'
        assert len(optAlphaArr) == len(alphas[i]), f'Incorrect number of alphas found'
        assert len(optBetaArr) == len(betas[i]), f'Incorrect number of betas found'

        assert not True in [f > forceErrorTolerance for f in (preserveOrderSort(forces[i], optForceArr) - forces[i])], f"Force magnitude exceeds tolerance for particle {i}."
        assert not True in [b > betaErrorTolerance for b in (preserveOrderSort(betas[i], optBetaArr, periodic=True) - betas[i])], f"Beta exceeds tolerance for particle {i}."
        assert not True in [a > alphaErrorTolerance for a in (preserveOrderSort(alphas[i], optAlphaArr, periodic=True) - alphas[i])], f"Alpha exceeds tolerance for particle {i}."



"""
Particle position tracking via convolution.
"""
import numpy as np
import cv2
import matplotlib.pyplot as plt

import numba
from scipy.fft import fft2, ifft2

from pepe.preprocess import circularMask, rectMask
from pepe.topology import findPeaks2D
from pepe.utils import lorentzian
from pepe.analysis import adjacencyMatrix

from lmfit import minimize, Parameters, fit_report

def convCircle(singleChannelFrame, radius, radiusTolerance=None, offscreenParticles=False, kernelBlurKernel=3, outlineOnly=False, outlineThickness=.05, negativeHalo=False, haloThickness=.03, negativeInside=False, peakDownsample=10, minPeakPrevalence=.1, intensitySoftmax=1.2, intensitySoftmin=.1, invert=False, allowOverlap=True, fitPeaks=True, debug=False):
    """
    Perform convolution circle detection on the provided image.

    A brief description of the technique is described below; for
    a more comprehensive discussion, see [1].

    The convolution method identifies circles in an image by convolving
    a kernel (representing what a particle should look like) with the
    image within which you would like to detect circles. This essentially
    finds the likelihood of any particle location being the center for
    such an object. By performing peak detection on this 2D quantity,
    we can identify the most likely places for particles to be.

    While this technique can be used for non-circular objects, this 
    implementation makes use of several simplifications/approximations
    that make it suitable only for radially symmetric objects.

    Tends to give more accurate and consistent results than the Hough
    technique (see `pepe.tracking.houghCircle()`), though evaluation
    times can be longer. When the radius of the circle is not well known
    (with no initial guess), the Hough method should be used. That being said,
    this implementation can fine tune the provided radius, but this requires
    the provided value to be reasonably close to the actual value (within 
    ~10 pixels is ideal).

    The ability to get subpixel precision by interpolating/fitting curves to the
    convolution is currently WIP, and currently the method can, at best, give results
    accurate to within a single pixel.

    
    Parameters
    ----------
 
    singleChannelFrame : np.uint8[H, W]
        The frame to perform circle detection on. Should not include multiple
        color channels. If it does, the channels will be averaged to form a
        grayscale image.

    radius : float or [float, float]
        The radius to look for, or min and max radii to constrain the circle
        detection. If a single value is provided, a range will be constructed
        around that value with the `radiusTolerance` parameter.

        Note that this method is more sensitive to the initial guess of radius
        than `pepe.tracking.houghCircle()`. Ideally, the radius should be within
        ~10 pixels of the true radius (or a range of radii should only be ~15 pixels wide).

        NOTE: Varying the radius is still under development, and likely does not work that great.

    radiusTolerance : float or None
        The half-width of the range of radii that will be tested for detecting particles.
        ie. the range will be calculated as `[radius + radiusTolerance, radius - radiusTolerance]`.

        If set to `None`, radius will be fixed at whatever value is provided via `radius` argument.

        NOTE: Varying the radius is still under development, and likely does not work that great.

    offscreenParticles : bool
        Whether or not the algorithm should look for particles whose centers are outside of
        the image domain (but some parts of the particles still show up in the image).

        This expands the padding provided to the fft/ifft calculations, so as to avoid cutting
        off what might be a potential particle. As a result, a value of True will increase computation times.

    outlineOnly : bool
        Whether to look for only a particle outline (True), or to look for a filled-in circle (False).

        See `outlineThickness`.

    outlineThickness : float or int
        Thickness of the particle outline in the kernel; only relevant if `outlineOnly=True`.

        If float, will be taken as fraction of the radius; if integer, will be taken as number
        of pixels.

    negativeHalo : bool
        Whether to surround the particle mask with a region of negative value, which
        further promotes identifying distinct objects as grains.

    haloThickness : float or int
        Thickness of the negative halo in the kernel; only relevant if `negativeHalo=True`.

        If float, will be taken as fraction of the radius; if integer, will be taken as number
        of pixels.

    negativeInside : bool
        Whether to fill the inside of the particle mask with a negative value, which 
        further promotes identifying outlines of circles.

        Has no effect unless `outlineOnly=True`.

    peakDownsample : int
        The factor by which to downsample the image when performing the initial peak
        detection. The final peak detection will always be performed at the original
        resolution, so this should not affect the accuracy of the detected circles,
        though it is possible that some particles in a densely packed system may
        not be detected. False

        Set to `1` or `None` to perform no downsampling, though this is not recommended,
        as the peak finding algorithm is O(N) where N is the total number of grid points.

    minPeakPrevalence : float
        The minimum prevalence (or persistence) of peaks that will be considered as
        particles. Should be a float in the range `[0, 1.]`, representing the percent
        of the total intensity data range that must be spanned by a feature.

        See `pepe.topology.findPeaks2D()` for more information.

    fitPeaks : bool
        Whether or not to fit each peak in the convolution with a lorentzian
        form and use that center as the location of the particle (instead of
        the simple maximum point).

    intensitySoftmax : float
        Value at which to cap the intensity values in the image, calculated as
        this value multipled by the mean of the image.

        Set to `None` to skip this preprocessing step.

    intensitySoftmin : float
        Zero-out all values that have an intensity below this factor times the mean of
        the image.

        Set to `None` to skip this preprocessing step.

    allowOverlap : bool
        Whether or not to allow detections of particles that overlap. If set to
        False, then the particle with the stronger signature will be kept
        
    debug : bool
        Whether or not to draw debug information on a plot. 

    invert : bool
        If particles appear dark on a bright background, this will invert the
        image (and convolution peak finding will go from looking for maxima
        to looking for minima).
   
    Returns
    -------

    centers : np.ndarray[N,2]
        The detected circles' centers.

    radii : np.ndarray[N]
        The detected circles' radii.


    References
    ----------

    [1] Franklin, S. V., & Shattuck, M. D. (Eds.). (2016). Handbook
    of Granular Materials. Chapter 2: Experimental Techniques. p.63-68.
    CRC Press. https://doi.org/10.1201/b19291

    """
    # General housekeeping of arguments, see if we need to parse anything
    if offscreenParticles:
        paddingFactor = 2.
    else:
        paddingFactor = 1.1

    if peakDownsample is None:
        peakDownsample = 1
 
    # Check if we are given a range of radii or just a single value
    possibleRadii = None

    if hasattr(radius, '__iter__'):
        if len(radius) == 2 and radius[0] < radius[1]:
           possibleRadii = np.arange(radius[0], radius[1]+1)
        else:
            raise Exception(f'Invalid radius provided to convCircle()!\n Expected scalar or [min, max], but got {radius}!')
    elif radiusTolerance is not None:
        # Use the tolerance
        possibleRadii = np.arange(radius - radiusTolerance, radius + radiusTolerance)
    else:
        # Just the single value
        possibleRadii = np.array([radius])

    # Start with just the mean of the possible radii options
    initialRadius = np.mean(possibleRadii)

    if singleChannelFrame.ndim > 2:
        imageArr = np.mean(singleChannelFrame, axis=-1)
    else:
        imageArr = np.copy(singleChannelFrame)
            
    if invert:
        imageArr = np.max(imageArr) - imageArr

    if intensitySoftmax is not None:
        softmax = np.mean(imageArr) * intensitySoftmax
        if not invert:
            imageArr[imageArr >= softmax] = softmax
        #else:
        #    imageArr[255 - imageArr >= softmax] = softmax

    if intensitySoftmin is not None:
        softmin = np.mean(imageArr) * intensitySoftmin
        if not invert:
            imageArr[imageArr <= softmin] = 0
        #else:
        #    imageArr[255 - imageArr <= softmin] = 0

    # Calculate convolution
    convArr = circularKernelFind(imageArr, initialRadius, fftPadding=int(initialRadius*paddingFactor),
                                 outlineOnly=outlineOnly, outlineThickness=outlineThickness,
                                 negativeHalo=negativeHalo, haloThickness=haloThickness,
                                 negativeInside=negativeInside, kernelBlurKernel=kernelBlurKernel, debug=debug)

    if debug:
        plt.show()

    # Downsample data by 5-10x and run peak detection
    downsampledConvArr = cv2.resize(cv2.blur(convArr, (peakDownsample,peakDownsample)), (0,0),
                                    fx=1/peakDownsample, fy=1/peakDownsample, interpolation=cv2.INTER_CUBIC)

    peakPositions, peakPrevalences = findPeaks2D(downsampledConvArr, minPeakPrevalence=minPeakPrevalence)

    lorentzPositions = np.zeros((len(peakPositions), 2))
    cleanedConvArr = np.zeros_like(convArr)

    # We have the option to fit lorentzian peaks to each maximum in the convolution
    # matrix, or we can just take the peak positions as they are
    if fitPeaks:
        def objectiveFunction(params, localImage):
            # Create a small grid across the area
            Y,X = np.ogrid[:localImage.shape[0], :localImage.shape[1]]
            Y += int(params["center_y"] - localImage.shape[0]/2)
            X += int(params["center_x"] - localImage.shape[1]/2)

            lorentzArr = lorentzian([Y,X], [params["center_y"], params["center_x"]], params["width"], params["amp"], params["offset"])
            
            return np.sum((localImage - lorentzArr)**2)

        # Now we fit an appropriate function to each peak to see if we can refine the center
        for i in range(len(peakPositions)):
            upsampledPosition = np.array([peakPositions[i][0]*peakDownsample, peakPositions[i][1]*peakDownsample])
            imagePadding = int(np.mean(possibleRadii))
            # Crop out the small area of the image around the peak
            localConv = convArr[max(upsampledPosition[0]-imagePadding, 0):min(upsampledPosition[0]+imagePadding,convArr.shape[0]-1), max(upsampledPosition[1]-imagePadding, 0):min(upsampledPosition[1]+imagePadding, convArr.shape[1]-1)]


            #plt.imshow(localConv)
            #plt.show()
            # Now setup a fit to this function
            params = Parameters()

            params.add('center_y', value=localConv.shape[0]/2, vary=True)
            params.add('center_x', value=localConv.shape[1]/2, vary=True)
            params.add('width', value=imagePadding/2, vary=False)
            params.add('amp', value=np.max(localConv)-np.min(localConv), vary=False)
            params.add('offset', value=np.min(localConv), vary=False)
        
            result = minimize(objectiveFunction, params, args=[localConv], method='nelder', max_nfev=10000)
            
            if result is not None:
                lorentzPositions[i,0] = upsampledPosition[0] - localConv.shape[0]/2 + result.params["center_y"]
                lorentzPositions[i,1] = upsampledPosition[1] - localConv.shape[1]/2 + result.params["center_x"]

                Y,X = np.ogrid[:convArr.shape[0], :convArr.shape[1]]
                cleanedConvArr += lorentzian([Y,X], lorentzPositions[i], result.params["width"], result.params["amp"], result.params["offset"])

            else:
                lorentzPositions[i] = upsampledPosition

    else:
        for i in range(len(peakPositions)):
            upsampledPosition = np.array([peakPositions[i][0]*peakDownsample, peakPositions[i][1]*peakDownsample])
            lorentzPositions[i] = upsampledPosition

    # Look around each peak to find the real peak in the full-resolution image
    # (only if we are actually downsampling)
    refinedPeakPositions = []
    refinedRadii = []

    if peakDownsample > 1:
        # To do this, we use a non-linear optimization scheme
        for i in range(len(peakPositions)):

            costArr = []

            def refineObjectiveFunction(params):
                kernelArr = genCircularKernel(np.array([params["y"].value, params["x"].value]), params["r"].value,
                                              imageArr.shape, kernelBlurKernel, outlineOnly,
                                              outlineThickness, negativeHalo, haloThickness, negativeInside)

                # We are minimizing this function, so we need the cost to be negative
                # (unless we have to invert the image)
                costArr.append((np.sum(kernelArr * imageArr) / np.sum(kernelArr)) * (int(invert)*2 - 1))
                return (np.sum(kernelArr * imageArr) / np.sqrt(np.sum(kernelArr))) * (int(invert)*2 - 1)

            params = Parameters()

            params.add('y', value=lorentzPositions[i][0], min=lorentzPositions[i][0]-imageArr.shape[0]*.02, max=lorentzPositions[i][0]+imageArr.shape[0]*.02)
            params.add('x', value=lorentzPositions[i][1], min=lorentzPositions[i][1]-imageArr.shape[1]*.02, max=lorentzPositions[i][1]+imageArr.shape[1]*.02)
            # May or may not want to vary the radius
            # Small tolerances on the radius are so that min != max
            params.add('r', value=np.mean(possibleRadii), vary=(len(possibleRadii) > 1), max=np.max(possibleRadii)-.01, min=np.min(possibleRadii)+.01)

            result = minimize(refineObjectiveFunction, params, method='powell', max_nfev=10000, options={"ftol": 1e-5, "xtol": 1e-5})

            #print(fit_report(result))

            if result is not None:
                refinedPeakPositions.append([result.params["y"].value, result.params["x"].value])
                refinedRadii.append(result.params["r"].value)
                
                #if debug:
                #    plt.plot(costArr)
                #    plt.show()

            else:
                refinedPeakPositions.append(lorentzPositions[i])
                refinedRadii.append(np.mean(possibleRadii))

    else:
        refinedPeakPositions = lorentzPositions
        refinedRadii = np.repeat(np.mean(possibleRadii), len(refinedPeakPositions))


    # Now we (optionally) remove overlapping particles
    # TODO: Allow overlap to be a float value, which allows overlaps
    # that do not exceed a certain threshold (eg. that fraction of the radius).
    if not allowOverlap:
        # Calculate the contact matrix
        # negative contact padding because we don't want to remove particles that are
        # good but just close to each other
        adjMat = adjacencyMatrix(refinedPeakPositions, refinedRadii, contactPadding=-int(np.mean(refinedRadii)/10)) - np.eye(len(refinedPeakPositions))

        # Locate all of the particles that are overlapping
        overlappingParticles = np.array([i for i in range(len(refinedPeakPositions)) if np.sum(adjMat[i,:]) > 0])

        # Now find how good of detections each of these is, and order them from worst to best.
        particleScores = np.zeros(len(overlappingParticles))
        for i in range(len(particleScores)):
            # Need to convert to int16 since we might have negative values
            pMask = genCircularKernel(refinedPeakPositions[i], refinedRadii[i],
                                      singleChannelFrame.shape, kernelBlurKernel, outlineOnly,
                                      outlineThickness, negativeHalo, haloThickness, negativeInside).astype(np.int16)

            if np.sum(pMask) > 0:
                # I square-root the mask sum so that the particles at the edge don't
                # always win out against actual particles
                # TODO: This still doesn't give priority to particles on screen;
                # need to figure out a way to do that
                particleScores[i] = np.sum(singleChannelFrame * pMask) / np.sqrt(np.sum(pMask))
            else:
                particleScores[i] = 0 

        # Sort according to worst scores
        order = np.argsort(particleScores)[::-1]
        overlappingParticles = overlappingParticles[order]

        # Now we start removing the particles with the worst scores until we no longer have overlap
        index = 0
        while np.sum(adjMat) > 0:
            adjMat[overlappingParticles[index],:] = 0
            adjMat[:,overlappingParticles[index]] = 0
            refinedPeakPositions[overlappingParticles[index]] = None
            refinedRadii[overlappingParticles[index]] = None

            index += 1

           
        # Now remove all of the None values we put in
        refinedPeakPositions = [rpp for rpp in refinedPeakPositions if rpp is not None]
        refinedRadii = [rr for rr in refinedRadii if rr is not None]


    # Debug option
    if debug:
        fig = plt.figure(figsize=(12,4))

        ax1 = fig.add_subplot(1, 3, 1, projection='3d')
        ax1.plot_surface(np.arange(convArr.shape[1]),
                        np.vstack(np.arange(convArr.shape[0])), convArr, cmap=plt.cm.jet)
        ax1.set_title('Convolution')

        ax2 = fig.add_subplot(1, 3, 2, projection='3d')
        ax2.plot_surface(np.arange(convArr.shape[1]),
                        np.vstack(np.arange(convArr.shape[0])), cleanedConvArr, cmap=plt.cm.jet)
        ax2.set_title('Cleaned Convolution')

        ax3 = fig.add_subplot(1, 3, 3)
        ax3.imshow(singleChannelFrame)
        ax3.set_title('Detected circles')
        for i in range(len(refinedPeakPositions)):
            c = plt.Circle(refinedPeakPositions[i][::-1], refinedRadii[i], color='red', fill=False, linewidth=1)
            ax3.add_artist(c)

        fig.tight_layout()

    return np.array(refinedPeakPositions), np.array(refinedRadii)


def circularKernelFind(singleChannelFrame, radius, fftPadding, kernelBlurKernel=3, outlineOnly=False, outlineThickness=.05, negativeHalo=False, haloThickness=.03, negativeInside=False, paddingValue=None, debug=False):
    """
    Calculate the convolution of a circular mask with an image,
    identifying likely locations of circles within the image. Adapted from
    method described in [1].

    To perform the convolution, we take the product of the fft of each the
    kernel and image, and then ifft the result to transform back to real space.
    The only peculiarity is that there can be some issues with the edges of image
    (especially if any circle centers are near the edge/off-screen) so we
    have to take care to properly pad the image and the kernel.

    Note that this method is not intended to be used as the front end for
    circle detection for real data. The method `pepe.tracking.convCircle()`
    makes use of this method to track circles, while also offering many more
    features. It is recommended to use that function unless there is a very
    specific purpose that it does not serve.

    Parameters
    ----------
    singleChannelFrame : np.ndarray[H,W]
        A single-channel image (grayscale or just one channel) within which we
        are looking to identify circular objects.

    radius : float
        The radius of circles to detect in the image.

    fftPadding : int
        The amount of padding to add to each side of the image, to prevent the
        fft from having issues. A good choice is usually 1.5 times the radius,
        though, less padding can be used if circles are not expected to often
        be located near the edges.

    kernelBlurKernel : int
        The kernel size to use to blur the kernel that will be convolved with the image, in
        pixels. A value of 0 or None will use the original kernel.

    outlineOnly : bool
        Whether to look for only a particle outline (True), or to look for a filled-in circle (False).

        See `outlineThickness`.

    outlineThickness : float or int
        Thickness of the particle outline in the kernel; only relevant if `outlineOnly=True`.

        If float, will be taken as fraction of the radius; if integer, will be taken as number
        of pixels.

    negativeHalo : bool
        Whether to surround the particle mask with a region of negative value, which
        further promotes identifying distinct objects as grains.

    haloThickness : float or int
        Thickness of the negative halo in the kernel; only relevant if `negativeHalo=True`.

        If float, will be taken as fraction of the radius; if integer, will be taken as number
        of pixels.

    negativeInside : bool
        Whether to fill the inside of the particle mask with a negative value, which 
        further promotes identifying outlines of circles.

        Has no effect unless `outlineOnly=True`.

    paddingValue : [float, `'sigmoid'`, or `None`]
        What value to fill in for the extra padding that is added during the fft/ifft.

        `None` leaves the value at 0.

        `'sigmoid'` will create a smooth-ish transition to 0.

        Any float value will be inserted into every point.

    debug : bool
        Whether or not to plot various quantities of the calculation for inspection
        at the end of the evaluation.

    Returns
    -------

    chiSqr : np.ndarray[H,W]
        The convolution of the kernel with the image. Larger values represent
        higher likelihood that there is a particle centered there.

    References
    ----------

    [1] Franklin, S. V., & Shattuck, M. D. (Eds.). (2016). Handbook
    of Granular Materials. Chapter 2: Experimental Techniques. p.64-68.
    CRC Press. https://doi.org/10.1201/b19291
    """

    if singleChannelFrame.ndim > 2:
        imageArr = np.mean(singleChannelFrame, axis=-1)
    else:
        imageArr = singleChannelFrame

    fftPadding = int(fftPadding)

    # Pad the proper image
    paddedImageArr = np.zeros((int(imageArr.shape[0] + 2*fftPadding), int(imageArr.shape[1] + 2*fftPadding)))
    paddedImageArr[fftPadding:-fftPadding,fftPadding:-fftPadding] = imageArr

    # Both dimensions will always have the same first crop index, but possibly a different second one
    cropBounds = [int(fftPadding + radius), int(-fftPadding + radius), int(-fftPadding + radius)]
    # If the fftpadding and the radius are exactly the same, we get some weird behavior
    # This happens because python can index negative values as that element from the end of
    # the array, but indexing -0 just gives 0, and so it gives you a (0,0) array.
    if cropBounds[1] >= 0:
        cropBounds[1] = None
        cropBounds[2] = None

    # We have a couple options for what values to put in the padding
    if paddingValue is None or paddingValue == 0:
        # Do nothing, and leave the value at 0
        pass

    elif paddingValue == 'sigmoid':
        # Sigmoid gives a smoothish decay from the image to 0 at the boundaries
        # Most of the constants here are just arbitrarily chosen to make things look good
        # Top
        paddedImageArr[:fftPadding] += np.multiply.outer(2/(1 + np.exp(np.linspace(0, 5, fftPadding))), paddedImageArr[fftPadding])[::-1,:] 
        # Bottom
        paddedImageArr[-fftPadding:] += np.multiply.outer(2/(1 + np.exp(np.linspace(0, 5, fftPadding))), paddedImageArr[-fftPadding-1])
        # Left
        paddedImageArr[:,:fftPadding] += np.multiply.outer(paddedImageArr[:,fftPadding], 2/(1 + np.exp(np.linspace(0, 5, fftPadding))))[:,::-1]
        # Right
        paddedImageArr[:,-fftPadding:] += np.multiply.outer(paddedImageArr[:,-fftPadding-1], 2/(1 + np.exp(np.linspace(0, 5, fftPadding))))

    else:
        # Just put a fixed value in
        paddedImageArr[:fftPadding] = paddingValue
        paddedImageArr[-fftPadding:] = paddingValue
        paddedImageArr[fftPadding:-fftPadding,:fftPadding] = paddingValue
        paddedImageArr[fftPadding:-fftPadding,-fftPadding:] = paddingValue

    center = np.array([radius,radius])
    # To be able to properly convolute the kernel with the image, they have to
    # be the same size, so just just put our kernel into an array the same size
    # as the image (though most entries will just be zero)
    # See function below for more information on generating this kernel
    kernelArr = genCircularKernel(center, radius, paddedImageArr.shape, kernelBlurKernel, outlineOnly,
                                 outlineThickness, negativeHalo, haloThickness, negativeInside)

    # First convolutional term
    convTerm1 = ifft2(fft2(paddedImageArr**2) * fft2(kernelArr))
    # Trim padding
    convTerm1 = convTerm1[cropBounds[0]:cropBounds[1],cropBounds[0]:cropBounds[2]]

    # Second term is pretty easy since we choose our weight function to be our
    # particle mask (and it just has 0s and 1s, so it is equal to it's square)
    convTerm2 = 2 * ifft2(fft2(paddedImageArr) * fft2(kernelArr))
    # Trim padding
    convTerm2 = convTerm2[cropBounds[0]:cropBounds[1],cropBounds[0]:cropBounds[2]]

    # For the normalization term, we want to sum all parts of the kernel that are
    # on screen (in the original image's frame) as we translate it across the image.
    originalImageMask = rectMask(paddedImageArr.shape, np.repeat(fftPadding, 2), np.array(imageArr.shape))[:,:,0]
    normalizationTerm = ifft2(fft2(originalImageMask) * fft2(kernelArr))
    # Trim padding
    normalizationTerm = np.real(normalizationTerm[cropBounds[0]:cropBounds[1],cropBounds[0]:cropBounds[2]])
    # Put everything together
    # Technically there is a +1 here, but that won't affect any minima, so we don't really care
    chiSqr = np.real((convTerm1 - convTerm2) / normalizationTerm)
    if np.min(chiSqr) < 0 and np.max(chiSqr) < 0:
        chiSqr = np.abs(chiSqr)

    if debug:
        fig, ax = plt.subplots(1, 4, figsize=(13,4))

        ax[0].imshow(kernelArr)
        ax[0].set_title('Kernel')

        ax[1].imshow(paddedImageArr)
        ax[1].set_title('Padded image')

        ax[2].imshow(normalizationTerm)
        ax[2].set_title('Normalization')

        ax[3].imshow(chiSqr)
        ax[3].set_title('Real-space convolution (Re)')

        fig.tight_layout()

    return chiSqr


def genCircularKernel(center, radius, imageShape, kernelBlurKernel=3, outlineOnly=False, outlineThickness=.05, negativeHalo=False, haloThickness=.03, negativeInside=False,):

    kernelArr = circularMask(imageShape, center, radius)[:,:,0].astype(np.float64)

    # If we only want the outline, we subtract another circular mask from the above one
    if outlineOnly:
        # Should be 2 if negative inside is true, 1 otherwise
        negativeInsideFactor = 1 + int(negativeInside)
        if outlineThickness < 1:
            innerRadius = np.ceil((1 - outlineThickness) * radius)
        else:
            innerRadius = radius - outlineThickness
        
        kernelArr = kernelArr - negativeInsideFactor * circularMask(imageShape, center, innerRadius)[:,:,0].astype(np.float64)

        if negativeHalo:
            if haloThickness < 1:
                haloRadius = radius + np.ceil((1 - haloThickness) * radius)
            else:
                haloRadius = radius + haloThickness

            kernelArr = 2*kernelArr - (circularMask(imageShape, center, haloRadius)[:,:,0].astype(np.float64) - circularMask(imageShape, center, radius)[:,:,0].astype(np.float64))

    elif negativeHalo:
        if haloThickness < 1:
            haloRadius = np.ceil((1 + haloThickness) * radius)
        else:
            haloRadius = radius + haloThickness

        kernelArr = 2*kernelArr - circularMask(imageShape, center, haloRadius)[:,:,0].astype(np.float64)

    if kernelBlurKernel is not None and kernelBlurKernel > 0:
        kernelArr = cv2.blur(kernelArr, (kernelBlurKernel,kernelBlurKernel))

    return kernelArr



"""
Decorator functions to simplify how kwargs are passed to pipline functions.
"""
from functools import wraps

def explicitKwargs():
    """
    Decorator that creates an attribute for the decorated function (`explicit_kwargs`)
    which is a list of the keyword arguments that are explicitly passed (and not just
    left to the default value).

    Adapted from:
    https://stackoverflow.com/questions/1408818/getting-the-keyword-arguments-actually-passed-to-a-python-method
    """
    def decorator(function):
        # Have to carry over the doc string, otherwise the documentation
        # won't generate properly
        @wraps(function)
        def inner(*args, **kwargs):
            inner.explicit_kwargs = kwargs.keys()
            return function(*args, **kwargs)
        return inner
    return decorator




import inspect
import numpy as np

from pepe.auto import forceSolve, forceSolveArgDTypes, excludedArgs

def test_forceSolve_DTypeCoverage():
    """
    Make sure that all of the dtypes for the arguments in
    pepe.auto.forceSolve are defined for the purpose of
    reading in values from a settings file.
    """
    args, counts= np.unique(list(inspect.signature(forceSolve).parameters.keys()) + list(forceSolveArgDTypes.keys()), return_counts=True) 
    missingArgs = [a for a in args[counts == 1] if a not in excludedArgs]
    
    assert len(missingArgs) == 0, f"Missing data types for args: {missingArgs}"




"""
Methods to solve for vector contact forces from a photoelastic image.
"""

import numpy as np

from pepe.preprocess import circularMask, rectMask, ellipticalMask, mergeMasks, upsample, downsample
from pepe.analysis import gSquared, adjacencyMatrix, testForceBalance
from pepe.simulate import genSyntheticResponse
from pepe.utils import outerSubtract

import numba
import copy

from lmfit import minimize, Parameters, fit_report
from lmfit.minimizer import AbortFitException

from scipy.signal import find_peaks

import matplotlib.pyplot as plt

# This dictionary defines the argument data types for
# forceOptimize(), which is used for reading in settings
# from a file
forceOptimizeArgDTypes = {"parametersToFit": list,
                      "method": str,
                      "maxEvals": int,
                      "forceBounds": tuple,
                      "betaBounds": tuple,
                      "alphaBounds": tuple,
                      "forceTolerance": float,
                      "betaTolerance": float,
                      "alphaTolerance": float,
                      "useTolerance": bool,
                      "returnOptResult": bool,
                      "allowAddForces": bool,
                      "allowRemoveForces": bool,
                      "minForceThreshold": float,
                      "contactMaskRadius": int,
                      "newBetaMinSeparation": float,
                      "newBetaG2Height": float,
                      "missingForceChiSqrThreshold": float,
                      "imageScaleFactor": float,
                      "localizeAlphaOptimization": bool,
                      "forceBalanceWeighting": float,
                      "debug": bool}

def initialForceSolve(photoelasticSingleChannel, centers, radii, fSigma, pxPerMeter, contactPadding=10, g2EdgePadding=1, contactG2Threshold=.1, contactMaskRadius=40, neighborEvaluations=4, boundaryMask=None, ignoreBoundary=True, brightfield=False, boundaryDetectionOptions={}, g2Cal=None):
    """
    Calculate the approximate forces on each particle based on the photoelastic response.
    No optimization/gradient descent is done in this method; this should be used either where
    rough estimates of the forces are fine, or as an initial condition for a fitting function.

    Heavily based on James Puckett's PhD thesis (2012) and Jonathan Kollmer's PeGS:
    https://github.com/jekollmer/PEGS

    Daniels, K. E., Kollmer, J. E., & Puckett, J. G. (2017). Photoelastic force
    measurements in granular materials. Review of Scientific Instruments, 88(5),
    051808. https://doi.org/10.1063/1.4983049

    Outline of the process:

    0. Track particles to get centers and radii (external)
    1. Calculate average gradient squared for each particle
    2. Find which particles are in contact with each other
    3. Determine if there are any boundary contacts (optional)
    4. Calculate the position of each (boundary or interparticle) force for each particle
    5. Use the average gradient from step 1 to estimate the magnitude of force at each position

    Intended to be used as the initial condition for optimization process; see
    `pepe.analysis.forceOptimize()` for more information.

    When significant noise is present in the images, this method will often way overestimate
    the magnitudes of forces (because it is based on the gradient squared). If the guesses
    are much too high, it can be better to just set all force magnitudes to a reasonable
    flat value, like `.1`.

    Parameters
    ----------

    photelasticSingleChannel : np.uint8[H,W]
        A single channel image in array form of the photoelastic response of particles.

    centers : np.ndarray[N,2] 
        A list of N centers of format [y, x].

    radii : np.ndarray[N]
        A list of N radii, corresponding to each particle center

    contactPadding : int
        Maximum difference between distance and sum of radii for which
        two particles will still be considered in contact.

    g2EdgePadding : int or float
        Number of pixels to ignore at the edge of each particle when calculating the average
        G^2. If float value < 1 is passed, gradient mask radius will be taken as that percent
        of the full particle radius. A value of 0 means no padding is included.

    contactG2Threshold : float
        The neighbor weight value under which edges will be removed from the network,
        as they would be considered to weak to represent anything physical. This will
        help remove particle neighbors that are only barely touching, but not transmitting
        any real force.

    contactMaskRadius : float
        The radius of the circular mask applied over a contact to estimate the magnitude
        of the force there. Should be smaller than the radius, usually 25% of the radius is fine.
        For higher forces, larger mask may be needed.

    neightborEvaluations : int
        How many of the closest points to find via the kd tree and test
        as potential contacts. For homogeneous circles, or approximately
        homogenous (< 2:1 size ratio), 4 should be plenty.

    boundaryMask : np.uint8[H,W]
        Optional mask removing uneccesary parts of the image, that is used
        to detect boundaries. Locations with a value of 0 will be assumed to be
        solid walls, that a particle could potentially be in contact with. If
        not provided, or set to None, system will be assumed to be in a rectangular
        box defined the most extreme top/bottom/left/right particles.

    ignoreBoundary : bool
        Whether to not attempt to find any contacts with the boundary. This will take effect
        regardless of whether a boundary mask is passed as a parameter.

    brightfield : bool
        Whether the photoelastic response is seen through a brightfield (True)
        polariscope or darkfield (False) polariscope.

    boundaryDetectionOptions : kwargs
        Dictionary of kwargs to be passed to the detectWallContact method, assuming
        boundaries are being included in the solver. If none are provided, relevant
        values are carried over from the kwargs for this method (`contactPadding`, 
        `g2EdgePadding`, `contactMaskRadius`, `contactG2Threshold`).

    g2Cal : float
        Gradient squared to force calibration value. For numerous calls of the
        function, it is recommended to calculate this value once externally,
        and pass it in. If left as None, will be recalculated internally.


    Returns
    -------

    forceGuessArr : np.object[N,F(N)]
        Triangular array of force magnitude guesses for N particles with F(N) forces each
        (could be different number of forces per particle).

    alphaGuessArr : np.object[N,F(N)]
        Triangular array of alpha angle guesses for N particles with F(N) forces each
        (could be different number of forces per particle).

    betaGuessArr : np.object[N,F(N)]
        Triangular array of beta angle guesses for N particles with F(N) forces each
        (could be different number of forces per particle).
    """

    # We are passed centers and radii
    numParticles = len(centers)
    # This should not include any channels, and should just be h x w
    imageSize = photoelasticSingleChannel.shape
   

    # Triangular arrays are a little wonky, but this type of initialization works
    betaGuessArr = [np.empty(0) for i in range(numParticles)]

    # Find neighbors
    # (This is the unweighted adjacency matrix, so a 1 if two particles are neighbors, 0 otherwise)
    # np.eye is the identity, which removes all diagonal "contacts" (a particle with itself)
    adjMat = adjacencyMatrix(centers, radii, contactPadding, neighborEvaluations) - np.eye(numParticles)

    # This includes duplicates, but that is ok because as long as we keep
    # the index order consistent, each force will show up only once for each particle
    # (but twice overall, which is what we want)
    contacts = np.transpose(np.where(adjMat))

    # Each pair of indices
    for p in contacts:
        betaGuessArr[p[0]] = np.append(betaGuessArr[p[0]], np.arctan2(centers[p[1]][1] - centers[p[0]][1], centers[p[1]][0] - centers[p[0]][0]))

    # Find potential wall contacts (optional)
    if not ignoreBoundary:
        # If no boundary is given, but ignoreBoundary = False, we just make a box around
        # our system
        if boundaryMask is None:
            # Working in image conventions, so top left corner is 0,0
            # So y value increases as you move downwards
            topLeftCorner = np.array([np.min(centers[:,0] - radii), np.min(centers[:,1] - radii)])
            dimensions = np.array([np.max(centers[:,0] + radii) - topLeftCorner[0],
                                   np.max(centers[:,1] + radii) - topLeftCorner[1]])

            boundaryMask = rectMask(photoelasticSingleChannel.shape, topLeftCorner, dimensions, channels=None)

        # We want to carry a few arguments over to the boundary detection
        # if they haven't been explicitly provided
        if not 'contactPadding' in boundaryDetectionOptions:
            boundaryDetectionOptions["contactPadding"] = contactPadding
        if not 'g2EdgePadding' in boundaryDetectionOptions:
            boundaryDetectionOptions["g2EdgePadding"] = g2EdgePadding
        if not 'contactMaskRadius' in boundaryDetectionOptions:
            boundaryDetectionOptions["contactMaskRadius"] = contactMaskRadius
        if not 'contactG2Threshold' in boundaryDetectionOptions:
            boundaryDetectionOptions["contactG2Threshold"] = contactG2Threshold

        # See detectWallContacts for more information
        numWallContacts, wallBetaArr, wallG2AvgArr = detectWallContacts(centers, radii, boundaryMask, photoelasticSingleChannel, **boundaryDetectionOptions)

        # Now merge these contacts with the interparticle ones
        for i in range(numParticles):
            for j in range(numWallContacts[i]):
                betaGuessArr[i] = np.append(betaGuessArr[i], wallBetaArr[i][j])

    # Alpha is very easy: we just assume all forces are radial
    # This creates an empty array the same shape as the beta one,
    # but with all of the values set to 0 (again, triangular arrays are wonky)
    alphaGuessArr = np.zeros(len(betaGuessArr), dtype='object')
    for i in range(len(alphaGuessArr)):
        alphaGuessArr[i] = np.zeros(len(betaGuessArr[i]))

    # Initialize force the same way, but this will end up with
    # actual values
    forceGuessArr = np.zeros(len(betaGuessArr), dtype='object')
    for i in range(len(forceGuessArr)):
        forceGuessArr[i] = np.zeros(len(betaGuessArr[i]))

    # Calculate G^2 of the image
    gSqr = gSquared(photoelasticSingleChannel)

    # Find G2 calibration value
    # Mean of the radii should be fine, though TODO would be a good idea to have options here
    if g2Cal is None:
        g2Cal = g2ForceCalibration(fSigma, np.mean(radii), pxPerMeter, brightfield=brightfield)

    # Figure out how much of the particle we will be calculating the g2 over
    if g2EdgePadding < 1. and g2EdgePadding > 0:
        g2MaskRadii = radii * g2EdgePadding
    elif g2EdgePadding >= 1:
        g2MaskRadii = radii - g2EdgePadding
    else:
        g2MaskRadii = radii

    for i in range(numParticles):
        g2Mask = circularMask(photoelasticSingleChannel.shape, centers[i], g2MaskRadii[i])[:,:,0]
 
        # This is the average g2 of the whole particle
        avgGSqr = np.sum(gSqr * g2Mask) / np.sum(g2Mask)

        # Now allocate this force to each contact
        contactG2Arr = np.zeros(len(forceGuessArr[i]))
        # We have to find the average gsqr at each contact
        for j in range(len(forceGuessArr[i])):
            contactMask = circularMask(photoelasticSingleChannel.shape,
                                       centers[i] + radii[i]*np.array([np.cos(betaGuessArr[i][j]), np.sin(betaGuessArr[i][j])]),
                                      contactMaskRadius)[:,:,0]
            # Now make sure it stays within the particle
            contactMask = 1 - np.int16((contactMask + g2Mask - 2) > 0)

            # This is average g2 of each contact
            contactG2Arr[j] = np.sum(gSqr * contactMask) / np.sum(contactMask)

        forceGuessArr[i][:] = avgGSqr * contactG2Arr / np.sum(contactG2Arr) / g2Cal

    # while (!solved)
    
        # Few iterations of least squares for each remaining particle
        # TODO

        # Find the ones with the least error, and perform a many more iterations
        # TODO

        # Apply force balance to those particles (optional)
        # TODO

        # Fix those contacts, and carry over the force to their neighbors
        # TODO

        # Remove fixed particles from the queue
        # TODO

    return forceGuessArr, alphaGuessArr, betaGuessArr


def forceOptimize(forceGuessArr, betaGuessArr, alphaGuessArr, radius, center, realImage, fSigma, pxPerMeter, brightfield, parametersToFit=['f', 'a'], method='nelder', maxEvals=300, forceBounds=(0, 5), betaBounds=(-np.pi, np.pi), alphaBounds=(-np.pi/2, np.pi/2), forceTolerance=.5, betaTolerance=.2, alphaTolerance=.1, useTolerance=True, returnOptResult=False, allowAddForces=True, allowRemoveForces=True, minForceThreshold=.01, contactMaskRadius=30, newBetaMinSeparation=.4, newBetaG2Height=.0005, missingForceChiSqrThreshold=2.1e8, imageScaleFactor=1, localizeAlphaOptimization=True, forceBalanceWeighting=.2, debug=False):
    """
    Optimize an initial guess for the forces acting on a particle using
    a nonlinear minimization function.

    Parameters
    ----------

    forceGuessArr : np.ndarray[Z]
        The initial guess for magnitudes of each force that is acting on the particle.

    betaGuessArr : np.ndarray[Z]
        The initial guess for the positional angles where each force is acting (in radians).

    alphaGuessArr : np.ndarray[Z]
        The initial guess for the directional angles for each force acting on the particle (in radians).

    radius : float
        The radius of the particle, in pixels.

    center : [int, int]
        The coordinates of the center, [y,x], of the particle.

    realImage : np.ndarray[H,W]
        The photoelastic channel that the optimizer will compare against. Can be a composite image
        include other particles, since a mask will be applied around the given particle.

    parametersToFit : ['f'[[, 'b'], 'a']]
        Which parameters should be allowed to be varied in the process of minimizing the function; 'f' for
        force magnitude, 'b' for beta positional angle, 'a' for alpha directional angle. Must include at least one.

        Can also specify to perform multiple optimizations on different variables by providing a 2D list of
        parameters keys (any of 'f', 'a', 'b') e.g. [ ['f', 'b'], ['f', 'a'] ] will first optimize the force
        and beta values, then optimize the force and alpha values. When performing multiple optimizations,
        the result of the previous one is used as initial condition for the following one, though the residual
        array includes information about all optimizations. 

        Values for `maxEvals`, `forceTol`, `betaTol`, `alphaTol`, and `method` can be provided as lists of
        acceptable values to specify different options for each optimization e.g. `method=['nelder', 'cobyla']`
        will use the Nelder-Mead optimization scheme on the first pass, and then COBYLA on the second one. If
        a single value is provided for any of these parameters, that will be used for all optimizations.

    method : str
        The method to use for optimization. See wiki page on Force Solver for more information on selecting
        an appropriate one. Generally any method from [lmfit](https://lmfit.github.io/lmfit-py/fitting.html#the-minimize-function)
        can be used, but the following tend to give the best results:

        - `lm` - Levenberg-Marquardt

        - `least_squares` - Trust region reflective

        - `nelder` - Nelder-Mead

        Can be provided as a list of values when performing multiple optimizations; see `parametersToFit`.

    maxEvals : int
        The maximum number of function evaluations before the minimizer exits. Convergence can occur before
        this number of evaluations, but process will be interrupted at this value regardless of whether
        or not the result has converged.

        Can be provided as a list of values when performing multiple optimizations; see `parametersToFit`.

    forceBounds : (float, float)
        The upper and lower limits for the values of force magnitudes that the minimizer can explore, 
        assuming that `useTolerance` is set to False.

    betaBounds : (float, float)
        The upper and lower limits for the values of beta that the minimizer can explore, 
        assuming that `useTolerance` is set to False.

    alphaBounds : (float, float)
        The upper and lower limits for the values of alpha that the minimizer can explore, 
        assuming that `useTolerance` is set to False.

    forceTolerance : float
        If `useTolerance` is set to True, this value will be used to calculate the bounds for each
        force individually, as the initial value minus this tolerance, and the initial value plus
        this tolerance.

        Can be provided as a list of values when performing multiple optimizations; see `parametersToFit`.

    betaTolerance : float
        If `useTolerance` is set to True, this value will be used to calculate the bounds for each
        beta individually, as the initial value minus this tolerance, and the initial value plus
        this tolerance.

        Can be provided as a list of values when performing multiple optimizations; see `parametersToFit`.

    alphaTolerance : float
        If `useTolerance` is set to True, this value will be used to calculate the bounds for each
        alpha individually, as the initial value minus this tolerance, and the initial value plus
        this tolerance.

        Can be provided as a list of values when performing multiple optimizations; see `parametersToFit`.

    useTolerance : bool
        Whether to calculate the bounds for each parameters as the initial value plus/minus a
        certain tolerance (True) or to use the set intervals provided via the `forceBounds`, 
        `betaBounds`, and `alphaBounds` parameters (False).

    returnOptResult : bool
        Whether to return the optimizer result object, which includes all parameters values and
        potentially uncertainties (True) or just the optimized values of the forces, betas, alphas,
        and residuals (False). Note that if multiple optimizations are being performed, only the
        final result object will be returned.
    
    allowAddForces : bool
        Whether or not to allow the optimization routine to attempt to detect missing forces,
        and add them in. This detection is done by examining the gradient squared of the 
        photoelastic image provided to the method (`realImage`). If a force is added, the
        last optimization routine will be run again, with the added ability to vary the
        contact position, beta (if this was not already the case).

    allowRemoveForces : bool
        Whether or not to allow the optimization routine to remove forces that have a
        magnitude below a certain threshold (see `minForceThreshold`).

    minForceThreshold : float
        The minimizer will automatically remove any forces whose magnitude is lower than
        this value between fittings and after the final one.

    contactMaskRadius : float
        The radius in pixels of the circular mask applied over contacts, either during localized
        alpha optimization (see `localizeAlphaOptimization`) or detection of missing
        contacts (see `allowAddForces`).
    
    newBetaMinSeparation : float
        The minimum separation for a potential new contact (see `allowAddForces`) to be
        considered as a real force to be added. Larger values can prevent doubling up
        on forces, when there is really only a single, extended contact point

    newBetaG2Height : float
        The threshold for average gradient squared value in a localized region around
        a potential new contact.

    missingForceChiSqrThreshold : float
        The minimum chi squared value (as calculated by the optimization routine; analagous to error)
        above which the method will attempt to detect missing forces.

        Found by experimentation.

    localizeAlphaOptimization : bool
        When optimizing for the incident angle (alpha) alone, the effect of varying
        this value can be very subtle across the entire particle. This parameter will have
        the largest effect in a small region around the contact. If this value is `True`, 
        optimizations only for alpha (done by passing `parametersToFit=['a']`, or similar)
        will mask the particle except for a small region around the contact.

    imageScaleFactor : int > 1 or positive float < 1
       Factor by which to scale the real image when optimizating. Integer larger than 1 will
       lead to the image being upsampled, and float less than 1 (but greater than 0) will lead
       to the image being downsampled.

       Must lead to an integer up- or downscale factor; for upscaling this is not an issue, but
       downscale factors may be rounded in not the exact inverse of an integer. eg. a value of `.35`
       will result in a downscale by a factor of 3, not 2.85.

    forceBalanceWeighting : float
        If set to some non-zero and non-None value, the optimization cost will include a
        contribution depending on how badly the ensemble of forces satisfy force balance.

    debug : bool
        Whether or not to print out status statements as the optimization is performed.
    """

    # If we are passed a 2d list of parametersToFit, that means we 
    # are to perform multiple minimizations, likely with different
    # parameters being optimized each time.

    # It is easier to set up this method to run for an arbitrary number
    # of optimizations, even though most of the time that number of
    # optimizations will be one

    # So if we aren't given a 2d list structure, we have to make one
    if type(parametersToFit[0]) is not list:
        parametersToFitList = [parametersToFit]
    else:
        parametersToFitList = parametersToFit

    numFits = len(parametersToFitList)

    # Different numbers of evaluations can be provided for each optimization
    # or just the one value can be used for all of them
    if type(maxEvals) is not list:
        maxEvalsList = [maxEvals for i in range(numFits)]
    else:
        maxEvalsList = maxEvals

    # The different minimizations can use different methods
    if type(method) is not list:
        methodList = [method for i in range(numFits)]
    else:
        methodList = method
    
    # Same deal for all of the tolerances (we don't do bound intervals, since tolerances
    # are the recommended way to handle parameters bounds)
    if type(forceTolerance) is not list:
        forceTolList = [forceTolerance for i in range(numFits)]
    else:
        forceTolList = forceTolerance

    if type(betaTolerance) is not list:
        betaTolList = [betaTolerance for i in range(numFits)]
    else:
        betaTolList = betaTolerance

    if type(alphaTolerance) is not list:
        alphaTolList = [alphaTolerance for i in range(numFits)]
    else:
        alphaTolList = alphaTolerance

    # Make sure everything is the same length
    assert len(parametersToFitList) == numFits, 'Invalid parametersToFit provided'
    assert len(maxEvalsList) == numFits, 'Invalid maxEvals provided'
    assert len(methodList) == numFits, 'Invalid method provided'
    assert len(forceTolList) == numFits, 'Invalid forceTol provided'
    assert len(betaTolList) == numFits, 'Invalid betaTol provided'
    assert len(alphaTolList) == numFits, 'Invalid alphaTol provided'

    # We need to keep track of the best set of parameters externally from
    # the optimization routine, in the case that the minimize() function
    # fails.
    global residuals, bestResidual, bestParams
    residuals = []
    bestResidual = 1e16 # Arbitrary large number that will be overwritten
    bestParams = None


    if imageScaleFactor > 1:
        # Make sure we have an integer factor
        realScaleFactor = int(imageScaleFactor)
        scaledImage = upsample(realImage, realScaleFactor)
    elif imageScaleFactor < 1:
        # Make sure we have the inverse of an integer factor
        realScaleFactor = 1/float(int(1/imageScaleFactor))
        scaledImage = downsample(realImage, int(1/imageScaleFactor))
    else:
        scaledImage = realImage
        realScaleFactor = 1

    scaledRadius = np.int64(radius * realScaleFactor)
    scaledCenter = np.array(center * realScaleFactor, dtype=np.float64)
    scaledPxPerMeter = np.int64(pxPerMeter * realScaleFactor)
    scaledContactMaskRadius = np.int64(contactMaskRadius * realScaleFactor)
    scaledParticleArea = scaledRadius**2 * np.pi

    # Setup our function based on what parameters we are fitting
    # We want to avoid any if statements within the function itself, since
    # that will be evaluated many many times.
    # lmfit has a nice setup in that you can denote whether a variable can be
    # changed or not, which means we don't actually have to change which variables
    # are passed to the function.
    def objectiveFunction(params, trueImage, z, radius, center, mask):
        global residuals, bestResidual, bestParams

        forceArr = np.array([params[f"f{j}"] for j in range(z)])
        betaArr = np.array([params[f"b{j}"] for j in range(z)])
        alphaArr = np.array([params[f"a{j}"] for j in range(z)])

        synImage = genSyntheticResponse(forceArr, alphaArr, betaArr, fSigma, scaledRadius, scaledPxPerMeter, brightfield, imageSize=scaledImage.shape, center=scaledCenter, mask=mask)

        # Error component for the response image matching up with the real one
        diffResidual = np.sum((synImage - trueImage)**2)
        # Error component for the forces being balanced
        # We scale by the particle area here, since the one above will be proportional to that
        # and we want the two to be somewhat comparable
        balanceResidual = np.sum(testForceBalance(forceArr, betaArr, alphaArr)**2) * scaledParticleArea * forceBalanceWeighting

        # Save residual for tracking error
        residuals.append(diffResidual + balanceResidual)

        # Save best configuration outside of minimization
        if residuals[-1] < bestResidual:
            bestResidual = residuals[-1]
            # Deep copy to make sure we aren't just assigning by reference
            bestParams = copy.deepcopy(params)
    
        return residuals[-1]


    # Mask our real image
    particleMask = circularMask(scaledImage.shape, scaledCenter, scaledRadius)[:,:,0]
    maskedImage = scaledImage * particleMask

    # Since we may be doing multiple fits, we want to set up the initial conditions
    # such that each fit uses the result of the previous one (and the first one
    # of course uses what is provided to the function)
    forceArr = np.array(forceGuessArr.copy())
    betaArr = np.array(betaGuessArr.copy())
    alphaArr = np.array(alphaGuessArr.copy())

    # Now that we have all of that bookkeeping done, we can actually get on
    # to doing the minimization
    result = None
    # Whether we should vary beta only for the next run, in case
    # we have added a force
    tempVaryBeta = False
    i = 0
    # We use a while loop here because we may want to repeat a fit if
    # we add a new force, which isn't possible in a for loop
    while i < numFits:

        z = len(forceArr)
       
        # If we lose all of our forces in the optimization,
        # the next optimization will fail.
        if z == 0:
            break

        # Out fitting parameters
        # if vary kwarg is false, that value won't be fit
        params = Parameters()

        for j in range(z):
            if useTolerance:
                # Have to make sure that certain values aren't allowed to go negative, but
                # otherwise the bounds are just the initial value +/- the tolerances
                params.add(f'f{j}', value=forceArr[j], vary='f' in parametersToFitList[i], min=max(forceArr[j]-forceTolList[i], 0), max=forceArr[j]+forceTolList[i])
                params.add(f'b{j}', value=betaArr[j], vary='b' in parametersToFitList[i] or tempVaryBeta, min=max(betaArr[j]-betaTolList[i], -np.pi), max=min(betaArr[j]+betaTolList[i], np.pi))
                params.add(f'a{j}', value=alphaArr[j], vary='a' in parametersToFitList[i], min=alphaArr[j]-alphaTolList[i], max=alphaArr[j]+alphaTolList[i])
            else:
                params.add(f'f{j}', value=forceArr[j], vary='f' in parametersToFitList[i], min=forceBounds[0], max=forceBounds[1])
                params.add(f'b{j}', value=betaArr[j], vary='b' in parametersToFitList[i] or tempVaryBeta, min=betaBounds[0], max=betaBounds[1])
                params.add(f'a{j}', value=alphaArr[j], vary='a' in parametersToFitList[i], min=alphaBounds[0], max=alphaBounds[1])

        # Reset the vary beta bool so we only do it this round
        tempVaryBeta = False

        # If alpha is the only parameter being adjusted, it can help to look only in a
        # localized region around each contact, since changing alpha has such a subtle
        # effect on the overall pattern.
        if localizeAlphaOptimization and 'a' in parametersToFitList[i] and not 'b' in parametersToFitList[i] and not 'f' in parametersToFitList[i]:
            # The *10 is to make sure that all points end up inside of the particle
            # eg. if you have two close contacts, and just look for places >= 2, we could
            # get a region that satisfies that condition, but isn't actually in the particle
            localizedMask = np.copy(particleMask) * 10

            for j in range(len(betaArr)):
                contactPoint = scaledCenter + scaledRadius * np.array([np.cos(betaArr[j]), np.sin(betaArr[j])])
                
                # Create a mask just over the small area inside of the particle
                localizedMask += circularMask(scaledImage.shape, contactPoint, scaledContactMaskRadius)[:,:,0]

            # >= 10 + 1 such that the points must be inside the
            # The value 10 is mostly arbitrary, but makes it very unlikely
            # that 10 separate contacts would overlap.
            localizedMask = (localizedMask >= 11).astype(np.uint8)
        else:
            # Otherwise we just set value to none, which will be handled by pepe.simulate.genSyntheticResponse()
            localizedMask = None

        # This is the callback function after each minimization step
        # There seems to be an error where sometimes the optimization
        # will raise an error if the max number of iterations is reached
        # (and sometimes it doesn't care) so we use this to make sure
        # we get the best set of parameters out, even if this error
        # comes up
        #def iterCallback(params, iter, resid, *args, **kwargs):
        #    if iter >= maxEvalsList[i]:

        # Just kidding, I'll just use a try and ignore the error, since I
        # am pretty sure that the params variable is constantly updated, not just
        # at the end.

        try:
            # Now do the optimization

            # The fatol kwarg is the minimum change between iterations before the
            # fit is considered to have converged. I just choose this based on an educated guess,
            # and it seems to work (worst case, you can just it really small and run the optimization
            # for the max number of evals every time).
            result = minimize(objectiveFunction, params,
                             args=(maskedImage, z, scaledRadius, scaledCenter, localizedMask),
                              method=methodList[i], max_nfev=maxEvalsList[i], nan_policy='omit', options={"fatol": 1e-2})

            # Copy over the new values of the forces, alphas, and betas
            for j in range(z):
                forceArr[j] = result.params[f"f{j}"] 
                betaArr[j] = result.params[f"b{j}"] 
                alphaArr[j] = result.params[f"a{j}"] 

        # If we run out of function evaluations, that not too huge a deal,
        # so we just grab the best value and move on. If we run into an actual error,
        # that shouldn't be caught here, since that might be serious, so we re-trhow it.
        except Exception as e:
            if not isinstance(e, AbortFitException):
                raise e

            if debug:
                print(e)

            # Otherwise, we take the last good value (since we kept track outside of
            # the optimization function)
            if bestParams is not None:
                for j in range(z):
                    forceArr[j] = bestParams[f"f{j}"] 
                    betaArr[j] = bestParams[f"b{j}"] 
                    alphaArr[j] = bestParams[f"a{j}"] 
            else:
                # If we don't have any good parameter values, just return the initial
                # guess. The arrays are initialized as this value, so we don't need
                # to do anything in this case.
                pass


        # ---------------------
        # Detect missing forces
        # ---------------------
        # If the code detects there is a missing force (no idea how yet)
        if result is not None and result.chisqr > missingForceChiSqrThreshold and allowAddForces:
            # We sweep around the edge of the particle to see if there
            # are any regions that look like they could have a force
            # (denoted by a particularly high g2 value, or rather a peak)
            testBetaCount = 30
            avgG2Arr = np.zeros(testBetaCount)
            newBetaArr = np.linspace(-np.pi, np.pi, testBetaCount)

            # Calculate all of the g2s around the edge of the particle
            gSqr = gSquared(scaledImage)
            for j in range(testBetaCount):
                contactPoint = scaledCenter + scaledRadius * np.array([np.cos(newBetaArr[j]), np.sin(newBetaArr[j])])
                
                # Create a mask just over the small area inside of the particle
                contactMask = circularMask(scaledImage.shape, contactPoint, scaledContactMaskRadius)[:,:,0]
                contactMask = (contactMask + particleMask) == 2

                avgG2Arr[j] = np.sum(contactMask * gSqr) / np.sum(contactMask)

            # Identify any peaks in the average g2s
            peakIndArr = find_peaks(avgG2Arr, height=newBetaG2Height)[0]
            peakIndArr = np.sort(peakIndArr)

            # Make sure that there aren't any artifacts of periodicity
            # Usually this isn't actually a problem, because the peak
            # finding algorithm requires a proper peak, which can only
            # be on one side (but we'll leave it here just in case)
            if np.arange(3).any() in peakIndArr and np.arange(len(avgG2Arr)-3, len(avgG2Arr)).any() in peakIndArr:
                # Remove last entry
                peakIndArr = peakIndArr[:-1]

            peakBetaArr = newBetaArr[peakIndArr]
            
            # Now we have a list of likely points, we need to see if our original
            # list is missing any of these.
            differenceArr = np.abs(np.subtract.outer(peakBetaArr, betaArr))

            # Check to see if there is a new peak that doesn't have
            # a previous force close to it
            for j in range(len(peakBetaArr)):
                if np.min(differenceArr[j]) > newBetaMinSeparation and np.max(differenceArr[j]) < 2*np.pi - newBetaMinSeparation:
                    # Add the new force
                    betaArr = np.append(betaArr, peakBetaArr[j])
                    forceArr = np.append(forceArr, .1) # Value isn't too imporant here
                    alphaArr = np.append(alphaArr, 0.)


            # If we have added a force, we should run the optimization again, and see if it improves
            if len(forceArr) > z:
                if debug:
                    print(f'Added {len(forceArr) - z} force(s).')
                # We also want to make sure we're allowed to vary beta on the next iteration
                tempVaryBeta = True

                # This skips the i += 1 at the end of the loop, and makes the optimization run again
                continue


        # ------------------------------------
        # Remove forces that don't do anything
        # ------------------------------------
        if len(forceArr[forceArr < minForceThreshold]) > 0 and allowRemoveForces:
            # Remove forces that aren't actually doing anything
            betaArr = betaArr[forceArr > minForceThreshold]
            alphaArr = alphaArr[forceArr > minForceThreshold]
            # This one has to be done last for the other indexing to work
            forceArr = forceArr[forceArr > minForceThreshold]
            if debug:
                print(f'Removed {z - len(forceArr)} force(s).')
       
        # Iterate (since we have a while not a for)
        i += 1

    if returnOptResult:
        return result

    else:
        return forceArr, betaArr, alphaArr, residuals


@numba.jit(nopython=True)
def g2ForceCalibration(fSigma, radius, pxPerMeter=1., g2Padding=1, alphaArr=np.array([0., 0.]), betaArr=np.array([0., -np.pi]), forceSteps=100, forceBounds=np.array([.01, 1.]), brightfield=True):
    """
    Use synthetic photoelastic response to fit the conversion constant between
    gradient squared value and force (in Newtons), assuming a linear relationship.

    Note that this computes the least squares of force (N) vs **average** gradient
    squared. The Matlab just uses the sum of gradient squared, but this is not invariant
    under changes of resolution, so I have opted to use the average. Because of this, a
    calibration value calculated in Matlab will **not** work here unless you divide out
    the number of points it is summed over first.

    Parameters
    ----------

    fSigma : float
        Stress optic coefficient, relating to material thickness, wavelength of light
        and other material property (C).

    radius : float
        Radius of the particle that is being simulated in pixels. If pxPerMeter is
        not provided (or set to 1), this value will be assumed to already have been converted to meters.

    contactMaskRadius : float
        Radius of the circular mask that is applied over each contact to find the average gradient
        squared.

    pxPerMeter : float
        The number of pixels per meter for the simulated image. If not provided, or set to 1, the radius
        value will be assumed to already have been converted to meters.

    g2Padding : int
        Number of pixels to ignore at the edge of the particle. We don't expect any boundary
        artifacts in our synthetic data, but we will eventually do this for the real data,
        so it is important to keep the size of the particles the same throughout.

    alphaArr : np.ndarray[Z]
        Array of angles representing force contact angles.

    betaArr : np.ndarray[Z]
        Array of angles representing force contact positions.

    forceSteps : int
        The number of points to use for fitting our line of g^2 vs. force.

    forceBounds : [float, float]
        The minimum and maximum value of force applied to calculate the calibration value.

    brightfield : bool
        Whether the intensity should be simulated as seen through a brightfield (True)
        polariscope or darkfield (False) polariscope.

    Returns
    -------

    g2ForceSlope : float
        Slope found via linear regression to convert average g^2 to force.
    """

    # The magnitude of the forces that will be acting at each step
    forceValues = np.linspace(forceBounds[0], forceBounds[1], forceSteps)
    gSqrAvgArr = np.zeros(forceSteps)

    imageSize = (np.int64(radius*2)+11, np.int64(radius*2)+11)
    center = np.array([imageSize[0]/2, imageSize[1]/2], dtype=np.float64)
    particleMask = circularMask(imageSize, center, radius - g2Padding)[:,:,0]

    # The contact mask is a circle placed over the edge of the particle where the force is applied
    #contactMask1 = circularMask(imageSize,
    #                            np.array([imageSize[0]/2 + radius*np.cos(betaArr[0]), imageSize[1]/2 + radius*np.sin(betaArr[0])]),
    #                            contactMaskRadius)[:,:,0]

    #contactMask2 = circularMask(imageSize,
    #                            np.array([imageSize[0]/2 + radius*np.cos(betaArr[1]), imageSize[1]/2 + radius*np.sin(betaArr[1])]),
    #                            contactMaskRadius)[:,:,0]

    # Get rid of the parts outside of the circle
    #contactMask1 = contactMask1 * particleMask
    #contactMask2 = contactMask2 * particleMask

    # Add them together
    #contactMask = contactMask1 + contactMask2

    # To divide out the number of points
    #numPoints = np.sum(contactMask)
    numPoints = np.sum(particleMask)

    for i in range(forceSteps):

        # Assume two forces acting on the particle with equal magnitude
        forceArr = np.array([forceValues[i], forceValues[i]])
       
        # Create a synthetic photoelastic response
        particleImg = genSyntheticResponse(forceArr, alphaArr, betaArr, fSigma, radius, pxPerMeter, brightfield, imageSize, center)

        # Calculate the gradient
        gSqr = gSquared(particleImg)

        # Multiply by the mask to avoid weird edge effects
        gSqrAvgArr[i] = np.sum(gSqr * particleMask) / numPoints

    # Now fit a straight line to the data
   
    # Create col vector of forces
    # Note that we multiply by 2 here
    # Since we are calculating g2 over the entire particle, we should
    # compare it to the force on the entire particle, not just one side.
    forceColMat = forceValues.reshape((forceSteps, 1)) * 2
    # Perform least squares
    solution = np.linalg.lstsq(forceColMat, gSqrAvgArr)

    return solution[0][0]


def g2ForceCalibrationDebug(fSigma, radius, pxPerMeter, alphaArr=np.array([0., 0.]), betaArr=np.array([0., -np.pi]), forceSteps=100, forceBounds=np.array([.01, 1.]), brightfield=True):
    # TODO: Also find the point at which the linear fit breaks down, so we can
    # return the max amount of force that can be converted using this method
    # before it stops working
    """
    Use synthetic photoelastic response to fit the conversion constant between
    gradient squared value and force (in Newtons), assuming a linear relationship.

    Returns the X and Y values that a line would be fit to, instead of the
    fit paramters themselves. See pepe.analysis.g2ForceCalibration().

    Note that this computes the least squares of force (N) vs **average** gradient
    squared. The Matlab just uses the sum of gradient squared, but this is not invariant
    under changes of resolution, so I have opted to use the average. Because of this, a
    calibration value calculated in Matlab will **not** work here unless you divide out
    the number of points it is summed over first.

    Parameters
    ----------

    fSigma : float
        Stress optic coefficient, relating to material thickness, wavelength of light
        and other material property (C).

    radius : float
        Radius of the particle that is being simulated in pixels. If pxPerMeter is
        not provided (or set to 1), this value will be assumed to already have been converted to meters.

    contactMaskRadius : float
        Radius of the circular mask that is applied over each contact to find the average gradient
        squared.

    pxPerMeter : float
        The number of pixels per meter for the simulated image. If not provided, or set to 1, the radius
        value will be assumed to already have been converted to meters.

    alphaArr : np.ndarray[Z]
        Array of angles representing force contact angles.

    betaArr : np.ndarray[Z]
        Array of angles representing force contact positions.

    forceSteps : int
        The number of points to use for fitting our line of g^2 vs. force.

    forceBounds : [float, float]
        The minimum and maximum value of force applied to calculate the calibration value.

    brightfield : bool
        Whether the intensity should be simulated as seen through a brightfield (True)
        polariscope or darkfield (False) polariscope.

    Returns
    -------

    forceArr : np.ndarray[forceSteps]
        Force applied (in Newtons) at each point

    g2AvgArr : np.ndarray[forceSteps]
        Resultant average gradient squared across the particle at each point.
    """

    # The magnitude of the forces that will be acting at each step
    forceValues = np.linspace(forceBounds[0], forceBounds[1], forceSteps)
    gSqrAvgArr = np.zeros(forceSteps)

    imageSize = np.array([np.int(radius*2.2), np.int(radius*2.2)])
    particleMask = circularMask(imageSize, imageSize/2, radius)[:,:,0]

    # The contact mask is a circle placed over the edge of the particle where the force is applied
    #contactMask1 = circularMask(imageSize,
    #                            np.array([imageSize[0]/2 + radius*np.cos(betaArr[0]), imageSize[1]/2 + radius*np.sin(betaArr[0])]),
    #                            contactMaskRadius)[:,:,0]
    
    #contactMask2 = circularMask(imageSize,
    #                            np.array([imageSize[0]/2 + radius*np.cos(betaArr[1]), imageSize[1]/2 + radius*np.sin(betaArr[1])]),
    #                            contactMaskRadius)[:,:,0]

    # Get rid of the parts outside of the circle
    #contactMask1 = contactMask1 * particleMask
    #contactMask2 = contactMask2 * particleMask

    # Add them together
    #contactMask = contactMask1 + contactMask2

    # To divide out the number of points
    #numPoints = np.sum(contactMask)
    numPoints = np.sum(particleMask)

    for i in range(forceSteps):

        # Assume two forces acting on the particle with equal magnitude
        forceArr = np.array([forceValues[i], forceValues[i]])
       
        # Create a synthetic photoelastic response
        particleImg = genSyntheticResponse(forceArr, alphaArr, betaArr, fSigma, radius, pxPerMeter, brightfield, imageSize, imageSize/2)

        # Calculate the gradient
        gSqr = gSquared(particleImg)

        # Multiply by the mask to avoid weird edge effects
        gSqrAvgArr[i] = np.sum(gSqr * particleMask) / numPoints

    # We multiply by 2 because that is the total force acting on the particle
    return 2*forceValues, gSqrAvgArr


#@numba.jit(nopython=True)
def detectWallContacts(centers, radii, boundaryMask, photoelasticSingleChannel=None, contactPadding=10, g2EdgePadding=.95, contactG2Threshold=1e-4, angleClusterThreshold=.2, contactMaskRadius=50, extendedContacts=False, maxContactExtent=.75):
    """
    Detect potential particle contacts with the wall.

    Parameters
    ----------

    centers : np.ndarray[N,2] 
        A list of N centers of format [y, x].

    radii : np.ndarray[N]
        A list of N radii, corresponding to each particle center

    boundaryMask : np.uint8[H,W]
        Mask removing uneccesary parts of the image, that is used
        to detect boundaries. Locations with a value of 0 will be assumed to be
        solid walls, that a particle could potentially be in contact with.

    photelasticSingleChannel : np.uint8[H,W]
        A single channel image in array form of the photoelastic response of particles.
        Is used to determine if a contact is actually force bearing or not. If not 
        provided, all potential contacts will be returned.

    contactPadding : int
        Maximum difference between distance between a particle's edge and the wall
        that will still be tested as a wall contact.

    g2EdgePadding : int or float
        Number of pixels to ignore at the edge of each particle when calculating the average
        G^2. If float value < 1 is passed, gradient mask radius will be taken as that percent
        of the full particle radius. A value of 0 means no padding is included.

    contactG2Threshold : float
        The neighbor weight value under which edges will be removed from the network,
        as they would be considered to weak to represent anything physical. This will
        help remove particle neighbors that are only barely touching, but not transmitting
        any real force.

    angleClusterThreshold : float
        The minimum difference in consecutive angle (beta) for which two wall contacts
        will be considered unique (and not merged into the same contact).

    contactMaskRadius : int
        The size of the circular mask that is used to determine average gradient squared
        value around detected contacts.

    extendedContacts : bool
        Whether or not to break up larger potential contacts in multiple. See `maxContactExtent`.

    maxContactExtent : float
        The maximum range of angles that can be included in a single cluster. If any particular
        cluster exceeds this value, it will be divided up into multiple new clusters.

    Returns
    -------

    numWallContacts : np.ndarray[N]
        The number of wall contacts for each particle (denoted as Z_i in other return values).
    
    betaArr : np.ndarray[N, Z_i]
        A list of angles for each particle in which there likely is a wall contact.

    contactG2Arr : np.ndarray[N, Z_i]
        A list of average gradient squared values for each wall contact, to be
        multiplied by the g2 calibration value to get the initial force magnitude guess.
    """

    # Figure out how much of the particle we will be calculating the g2 over
    g2MaskRadii = radii.astype(np.float64)
    if g2EdgePadding < 1. and g2EdgePadding > 0:
        g2MaskRadii = (radii.astype(np.float64) * g2EdgePadding)
    elif g2EdgePadding >= 1:
        g2MaskRadii = (radii - g2EdgePadding).astype(np.float64)

    # Things that will be returned
    numWallContacts = np.zeros(len(centers), dtype=np.int16)
    betaArr = []
    contactG2Arr = []

    # Iterate over each particle
    for i in range(len(centers)):

        # Create a mask slightly larger than the particle
        wallContactMask = circularMask(boundaryMask.shape, centers[i], radii[i] + contactPadding)

        # Find the overlap (similarity) of this enlarged particle mask and the boundary mask
        invBoundaryMask = 1 - boundaryMask
        similarity = np.floor((invBoundaryMask + wallContactMask).astype(np.double)/2).astype(np.uint8)[:,:,0]

        # Determine if there is a chance of a wall contact
        if np.sum(similarity) > 0:
            # ------------------------------
            # Find every point that overlaps
            # ------------------------------

            # Normally we could just do this:
            #points = np.transpose(np.where(cMask > 0))
            # but numba doesn't quite like this way, so we have to be 
            # a little more creative

            whereIndices = np.where(similarity > 0)
            points = np.zeros((len(whereIndices[0]), 2), dtype=np.int16)
            # There is a chance that these indices are backwards, but
            # because we have rotational symmetry, it doesn't really matter...
            # BUT if there is ever some weird anisotropy bug or something,
            # try switching these indices
            points[:,0] = whereIndices[0]
            points[:,1] = whereIndices[1]

            # ------------------------------
            # Cluster them based on position
            # ------------------------------

            # Convert to angles
            angles = np.arctan2(points[:,1] - centers[i][1], points[:,0] - centers[i][0])
            
            # Sort the angles, since they may not exactly be in order
            sortedIndices = np.argsort(angles)
            angles = np.sort(angles)

            # To apply it to the original points, we would have to reverse the sorting
            # we did earlier, or much easier, just sort the original points the same way
            points = points[sortedIndices]

            # Calculate the difference between angles
            # There should be a jump once we move to a new cluster
            dAngle = angles[1:] - angles[:-1]

            # Add 1 to these, because we started at 1
            # These are the indices of places that have large jumps in angle
            clusterSeparation = np.where(dAngle > angleClusterThreshold)[0].astype(np.int16) + 1

            # Generate an array with labels for each angle
            numClusters = len(clusterSeparation) + 1
            labels = np.zeros(len(points))
            startingIndex = 0

            for j in range(numClusters-1):
                labels[startingIndex:clusterSeparation[j]] = j
                startingIndex = clusterSeparation[j]
            # And the final cluster
            labels[startingIndex:] = numClusters-1

            # Check to see if we have any artifacts of the periodicity of the angle
            # (and fix them by setting the last cluster to be equal to the first one)
            #print(angles[0], angles[-1])
            if numClusters > 1 and abs(angles[0] - angles[-1]) < angleClusterThreshold or abs(abs(angles[0] - angles[-1]) - 2*np.pi) < angleClusterThreshold:
                labels[labels == np.max(labels)] = 0
                numClusters -= 1

            # --------------------------------------
            # Calculate the centroid of each cluster
            # --------------------------------------
            clusterCentroids = np.zeros((numClusters, 2))
            for j in range(numClusters):
                clusterCentroids[j] = [np.mean(points[labels == j][:,0]), np.mean(points[labels == j][:,1])]
 

            # --------------------------------------------------------------
            # Calculate the angle with respect to the particle center (beta)
            # --------------------------------------------------------------
            clusterBetas = np.zeros(numClusters)
            for j in range(numClusters):
                clusterBetas[j] = np.arctan2(clusterCentroids[j,1] - centers[i][1], clusterCentroids[j,0] - centers[i][0])

            # -------------------------------------
            # Divide up big clusters (if necessary)
            # -------------------------------------
            if extendedContacts:
                newBetas = np.zeros(0)
                for j in range(numClusters):
                    # First, calculate the extent of the cluster
                    # This isn't as simple as subtract max from min, because
                    # of the periodicity, so the most reliable method is as follows

                    # Locate every unique angle in this cluster (in order)
                    uniqueBetas = np.sort(np.unique(angles[labels == j]))

                    # If you only have 1 beta, then clearly we don't need to divide
                    # this cluster up
                    if len(uniqueBetas) < 2:
                        newBetas = np.append(newBetas, clusterBetas[j])
                        continue

                    clusterBounds = np.array([np.max(np.array([uniqueBetas[0], uniqueBetas[-1]])), np.min(np.array([uniqueBetas[0], uniqueBetas[-1]]))])
                    clusterExtent = clusterBounds[0] - clusterBounds[1]

                    # This is usually a good way to identify that the region
                    # passes across the top of the circle
                    if (clusterExtent < .01 or 2*np.pi - clusterExtent < .01) and np.min(uniqueBetas) < 0 and np.max(uniqueBetas) > 0:
                    #if (clusterBetas[j] < clusterBounds[0] and clusterBetas[j] > clusterBounds[1]):
                        clusterBounds = [np.max(uniqueBetas[uniqueBetas < 0]), np.min(uniqueBetas[uniqueBetas > 0])]
                        clusterExtent = 2*np.pi - (clusterBounds[1] - clusterBounds[0])

                    if clusterExtent > maxContactExtent:
                        numNewClusters = np.int16(np.ceil(clusterExtent / maxContactExtent))
                        dBeta = clusterExtent/numNewClusters
                        newBetas = np.append(newBetas, np.linspace(clusterBetas[j] + clusterExtent/2., clusterBetas[j] - clusterExtent/2., numNewClusters))
                    else:
                        newBetas = np.append(newBetas, clusterBetas[j])
                        
                #print(newBetas)

                clusterBetas = newBetas.copy()
                numClusters = len(clusterBetas)

            # Now we want to recalculate our centroids, since there are
            # potentially some new ones
            clusterCentroids = np.zeros((numClusters, 2))
            for j in range(numClusters):
                clusterCentroids[j] = centers[i] + radii[i] * np.array([np.cos(clusterBetas[j]), np.sin(clusterBetas[j])])

            # --------------------------------------------------------------------------
            # Apply a mask to get the magnitude of the average g2 value for that contact
            # --------------------------------------------------------------------------
            # We only do this if we are provided a photoelastic image
            clusterAvgG2 = np.zeros(numClusters)

            if photoelasticSingleChannel is not None:
                # Calculate G2
                gSqr = gSquared(photoelasticSingleChannel)

                for j in range(numClusters):
                    # Create an elliptical mask around the area the photoelastic response would
                    # be in. This is actually more a half ellipse, since we want a reasonably
                    # wide area around the contact itself, and only a small bit of the center
                    # of the particle.
                    #majorAxisDir = clusterCentroids[j] - centers[i]
                    #majorAxisDir /= np.sqrt(majorAxisDir[0]**2 + majorAxisDir[1]**2)
                    #ellipseE radii[i] - g2EdgePaddingndpoint = majorAxisDir * 2 *radii[i] + centers[i]
                    #contactMask = ellipticalMask(photoelasticSingleChannel.shape, centers[i], ellipseEndpoint, radii[i]/2)

                    # Just kidding, a circular mask seems to work better
                    contactMask = circularMask(photoelasticSingleChannel.shape, clusterCentroids[j], contactMaskRadius)

                    # Now cut off the part outside of the particle
                    contactMask = (contactMask + circularMask(photoelasticSingleChannel.shape, centers[i], g2MaskRadii[i]))[:,:,0] == 2

                    # Calculate average G2
                    clusterAvgG2[j] = np.sum(contactMask * gSqr) / np.sum(contactMask)


            # ---------------------------
            # Save all of the information
            # ---------------------------            
            numWallContacts[i] = len(clusterBetas)
            betaArr.append(clusterBetas)
            contactG2Arr.append(clusterAvgG2)

        else:
            # No possible wall contacts
            # numWallContacts is initialized at 0, so don't need to set it
            # Numba doesn't allow untyped lists, so this is a little trick
            # to get an empty list into this array
            betaArr.append(np.array([np.double(x) for x in range(0)]))
            contactG2Arr.append(np.array([np.double(x) for x in range(0)]))
            pass


    return numWallContacts, betaArr, contactG2Arr




"""
Common functional forms to fit to.
"""
import numpy as np

def cone(p, center, halfAngle, offset, roundness):
    """
    A mix between a cone and a gaussian.

    Parameters
    ----------

    p : [float, float]
        The coordinates [y,x] of the point to compute the height on the cone.

    center : [float, float]
        The coordinates [y,x] of the center of the cone.

    halfAngle : float
        The half angle (in radians) of the cone.

    offset : float
        The z coordinate of the tip of the cone.

    roundness : float
        The factor of roundness for the tip of the cone; 0 is a perfectly
        sharp cone, and values greater than 0 denote increasing roundness.

    Returns
    -------

    z : float
        The z position at the point `p` on the cone.
    """
    d = np.sqrt((center[0] - p[0])**2 + (center[1] - p[1])**2)

    # Otherwise we will have /0 error
    if d == 0:
        return offset

    return offset - d/np.tan(halfAngle) * np.exp(-(roundness)**2/d)


def lorentzian(p, center, width, amp, offset):
    r"""
    The 2D Lorentzian function, evaluated at a single point.

    Parameters
    ----------

    p : [float, float]
        The coordinates [y,x] of the point to compute the height on the Lorentzian.

    center : [float, float]
        The coordinates [y,x] of the center of the Lorentzian.
    
    width : float
        The width of the Lorentzian.

    amp : float
        The amplitude of the Lorentzian.

    offset : float
        The z coordinate of the tip of the Lorentzian.

    Returns
    -------

    z : float
        The z position at the point `p` on the Lorentzian.
    """
    return amp * 0.5 * width / ( ((center[0] - p[0])**2 + (center[1] - p[1])**2) + 0.25 * width**2) + offset



import numpy as np

from pepe.utils import *

"""
Parse.py
"""

def test_parseList_StrList():
    assert parseList("['a', 'b', 'c']", str) == ['a', 'b', 'c']

def test_parseList_IntList():
    assert parseList("[1, 2, 3, 4]", int) == [1, 2, 3, 4]

def test_parseList_NestedStrList(): 
    assert parseList("[['a', 'b'], ['b', 'c'], ['c']]") == [['a', 'b'], ['b', 'c'], ['c']]

def test_parseList_NestedIntList():
    assert parseList("[[1, 2], [3], [4, 5]]") == [[1, 2], [3], [4, 5]]

"""
Sort.py
"""

def test_preserveOrderArgsort_OneDim():
    # len(new) == len(old)
    new = [2.1, 1.1, 4.1, 5.1, 3.1]
    old = [1, 3, 2, 4, 5]
    assert preserveOrderArgsort(old, new) == [1, 4, 0, 2, 3]

def test_preserveOrderArgsort_OneDimPad():
    # len(new) < len(old), pad
    new = [2.1, 1.1, 4.1, 5.1]
    old = [1, 3, 2, 4, 5]
    assert preserveOrderArgsort(old, new, padMissingValues=True) == [1, None, 0, 2, 3]

def test_preserveOrderArgsort_OneDimPad_2():
    # len(new) > len(old), pad
    new = [2.1, 1.1, 4.1, 5.1]
    old = [2, 4, 5]
    assert preserveOrderArgsort(old, new, padMissingValues=True) == [0, 2, 3, 1]

def test_preserveOrderArgsort_OneDimNoPad():
    # len(new) < len(old), no pad
    new = [2.1, 1.1, 4.1, 5.1]
    old = [1, 3, 2, 4, 5]
    assert preserveOrderArgsort(old, new, padMissingValues=False) == [1, 0, 2, 3]

def test_preserveOrderArgsort_OneDimNoPad_2():
    # len(new) > len(old), no pad
    new = [2.1, 1.1, 4.1, 5.1]
    old = [2, 4, 5]
    assert preserveOrderArgsort(old, new, padMissingValues=False) == [0, 2, 3, 1]

def test_preserveOrderArgsort_OneDimPadDistance():
    # len(new) < len(old), pad, distance
    new = [2.1, 1.1, 4.1]
    old = [1, 3, 2, 4, 5]
    assert preserveOrderArgsort(old, new, padMissingValues=True, maxDistance=.5) == [1, None, 0, 2, None]

def test_preserveOrderArgsort_TwoDim():
    # len(new) == len(old)
    new = [[1, 2], [3, 3], [5, 4]]
    old = [[.9, 2], [4.9, 4.1], [2.9, 2.9]]
    assert preserveOrderArgsort(old, new) == [0, 2, 1]

def test_preserveOrderArgsort_TwoDimPad():
    # len(new) < len(old), pad
    new = [[3, 3], [5, 4]]
    old = [[.9, 2], [4.9, 4.1], [2.9, 2.9]]
    assert preserveOrderArgsort(old, new, padMissingValues=True) == [None, 1, 0]

def test_preserveOrderArgsort_TwoDimNoPad():
    # len(new) < len(old), no pad
    new = [[3, 3], [5, 4]]
    old = [[.9, 2], [4.9, 4.1], [2.9, 2.9]]
    assert preserveOrderArgsort(old, new, padMissingValues=False) == [1, 0]

def test_preserveOrderArgsort_TwoDimPadDistance():
    # len(new) > len(old), pad, distance
    new = [[3, 3], [5, 4], [7, 7], [9, 10], [11, 11]]
    old = [[.9, 2], [4.9, 4.1], [2.9, 2.9], [10.9, 10.9]]
    assert preserveOrderArgsort(old, new, padMissingValues=True, maxDistance=.5) == [None, 1, 0, 4, 2, 3]

def test_preserveOrderArgsort_TwoDimNanFill():
    # len(new) < len(old), pad, distance, fill
    new = [[3, 3], [5, 4], [9, 9]]
    old = [[.9, 2], [np.nan, np.nan], [4.9, 4.1], [2.9, 2.9]]
    assert preserveOrderArgsort(old, new, padMissingValues=True, maxDistance=.5, fillNanSpots=True) == [None, 2, 1, 0]

def test_preserveOrderArgsort_TwoDimNanNoFill():
    # len(new) < len(old), pad, distance, no fill
    new = [[3, 3], [5, 4], [9, 9]]
    old = [[.9, 2], [np.nan, np.nan], [4.9, 4.1], [2.9, 2.9]]
    assert preserveOrderArgsort(old, new, padMissingValues=True, maxDistance=.5, fillNanSpots=False) == [None, None, 1, 0, 2]


"""
Outer.py
"""

def test_outerSubtract_2x2():
    a = np.array([1, 2])
    b = np.array([2, 3])
    assert np.sum(outerSubtract(a, b) - np.array([[-1, -2], [0, -1]])) == 0

def test_outerSubtract_3x3():
    a = np.array([1, 2, 3])
    b = np.array([1, 2, 3])
    assert np.sum(outerSubtract(a, b) - np.array([[0, -1, -2], [1, 0, -1], [2, 1, 0]])) == 0


"""
RectArr.py
"""

def test_rectArr_FakeData():
    cArr = [[[2, 1], [9, 15]], # a, b
           [[10, 15], [3, 2]], # b, a
           [[4, 2], [10, 14]]] # a, b

    rArr = [[5, 5], [6, 5], [5, 6]]

    fArr = [[[.1, .5], [.3]],
           [[.4, .1], [.45, .05]],
           [[.43], [.46, .2]]]

    aArr = [[[.02, .01], [.2]],
           [[.25, .01], [.05, .04]],
           [[.03], [.26, .02]]]

    bArr = [[[0, 3], [1.5]],
           [[1.52, 0], [2.99, 0.02]],
           [[2.95], [1.55, .02]]]

    rfArr, raArr, rbArr, rcArr, rrArr = rectangularizeForceArrays(fArr, aArr, bArr, cArr, rArr)

    assert len(rcArr) == 2
    assert np.sum(rcArr[0] - np.array([[2, 1], [3, 2], [4, 2]])) == 0
    assert np.sum(rcArr[1] - np.array([[9, 15], [10, 15], [10, 14]])) == 0

    assert np.sum(rrArr[0] - np.array([5, 5, 5])) == 0
    assert np.sum(rrArr[1] - np.array([5, 6, 6])) == 0



"""
Method to solve D2min field between two different particle states,
and calculate the von Mises strain.

Copied directly from my standalone repo:
https://github.com/Jfeatherstone/D2min

Originally forked from:
https://github.com/Binxu-Stack/D2min
created by Bin Xu.

Modifications include more documentation, change of notation,
and change from bond-focused calculation to center-focused
calculation (see code for more details).

Intended for use within other projects, so CLI access has been
removed.
"""

import numpy as np
from numpy.linalg import inv

# For calculating which neighbors are within a certain distance
from sklearn.neighbors import KDTree

def calculateD2Min(initialCenters, finalCenters, refParticleIndex=0, interactionRadius=None, interactionNeighbors=None, normalize=True):
    """
    Calculate the d2min between an initial and final state of particles;
    a measure of how non-affine the transformation is, as originally described
    in [1].

    
    Parameters
    ----------

    initialCenters : np.ndarray[N,d] or list
        List of initial center points of N particles in d dimensions.

    finalCenters :  np.ndarray[N,d] or list
        List of final center points of N particles in d dimensions. Must be
        in the same order as initialCenters.

    refParticleIndex : int
        The index of the particle to treat as the reference particle (r_0 in Falk & Langer
        1998 eq. 2.11). If set to None, will calculate the D2min N times using each particle
        as the reference (and return types will have an extra first dimension N). Simiarly, can
        be a list of indices for which to calculate as the refernce indices.

    interactionRadius : float
        The maximum distance between particles that can be considered neighbors. Recommended to
        be set to around 1.1 - 1.5 times the mean particle radius of the system.

        If set to None, all other particles in the system will be considered neighbors. See interactionNeighbors
        for specifying a fixed number of neighbors. In the case that neither a radius or number of
        neighbors are specified, calculation will default to using all other particles as neighbors.

    interactionNeighbors : int
        As opposed to using an interactionRadius to define neighbors, a fixed number of neighbors can
        be specified here. This number of neighbors will be found using a kd-tree for the reference point(s).

        In the case that neither a radius or number of neighbors are specified, calculation will default
        to using all other particles as neighbors.

    normalize : bool
        Whether to divide the d2min by the number of neighbors used to calculate it (True) or not (False).
        For heterogeneous systems where the number of neighbors can vary significantly, recommend to set True.
        Will make little difference if a fixed number of neighbors (see interactionNeigbors) are used.

    Returns
    -------

    d2min : float
        The minimum value of D2 for the transition from the initial to final state. Units
        are a squared distance dimensionally the same as the initial and final centers (likely
        a pixel^2 value if tracked from images). Changing units to particle diameters afterwards
        may be necessary.

    epsilon : numpy.ndarray[d,d]
        The uniform strain tensor that minimizes D2; equation 2.14 in [1].


    In the case that `refParticleIndex=None`, the return will instead be a tuple of numpy arrays
    containing the same information but for every particle:

    d2minArr : np.ndarray[N]
        The minimum value of D2 for the transition from the initial to final state for
        every possible configuration.

    epsilonArr : np.ndarray[N,d,d]
        The uniform strain tensor that minimizes D2 for every possible configuration


    Examples
    --------

    See `test` folder in [standalone repository](https://github.com/Jfeatherstone/D2min).

 
    References
    ----------    

    [1] Falk, M. L., & Langer, J. S. (1998). Dynamics of viscoplastic deformation in amorphous solids. Physical Review E, 57(6), 7192–7205.
    [https://doi.org/10.1103/PhysRevE.57.7192](https://doi.org/10.1103/PhysRevE.57.7192)

    """

    # The number of particles and spatial dimension
    N, d = np.shape(initialCenters)

    # In the case that a single reference particle is defined, we just calculate exactly as described in the paper
    if not isinstance(refParticleIndex, list) and not isinstance(refParticleIndex, np.ndarray) and refParticleIndex != None:

        # Determine which particles are neighbors using the parameters supplied on the function call
        initialNeighbors = None
        finalNeighbors = None

        # If an interaction radius is supplied, we have to find the particles that are closest
        # in the initial state using a kd-tree
        if interactionRadius != None:
            kdTree = KDTree(initialCenters)
            ind = kdTree.query_radius([initialCenters[refParticleIndex]], interactionRadius)
            
            # We ignore the first element since it will always be the particle itself
            # And we have to sort we don't mess up the order
            ind = np.sort(ind[0][1:])

            # Make sure we actually found some particles
            if len(ind) == 0:
                print('Warning: no neighbors found within supplied interaction radius! Defaulting to all other particles')
            else:
                initialNeighbors = initialCenters[ind]
                finalNeighbors = finalCenters[ind]    

        # If a fixed number of neighbors is provided instead, we find those particles again with
        # a kd-tree
        elif interactionNeighbors != None:
            kdTree = KDTree(initialCenters)
            # Make sure we don't look for more particles than are in our system
            dist, ind = kdTree.query([initialCenters[refParticleIndex]], min(interactionNeighbors+1, N))
            
            # We ignore the first element since it will always be the particle itself
            # And we have to sort we don't mess up the order
            ind = np.sort(ind[0][1:])
            
            initialNeighbors = initialCenters[ind]
            finalNeighbors = finalCenters[ind]    

        # If no information is supplied, or we ran into issues, use every other particle
        if not isinstance(initialNeighbors, list) and not isinstance(initialNeighbors, np.ndarray):
            initialNeighbors = initialCenters[np.arange(N) != refParticleIndex]
            finalNeighbors = finalCenters[np.arange(N) != refParticleIndex]

        # Now onto the actual D2min calculation
        # Bin's original code defined the differences between centers in
        # Falk & Langer eq. 2.11 - 2.13 as "bonds"

        # We first calculate these bonds using our reference particle index
        # Do this by subtracting the ref particle center from the center of every other particle
        # Note that you could technically leave in the ref bond, since it will be 0 and not contribute,
        # but it is cleaner to just remove it

        initialBonds = initialNeighbors - initialCenters[refParticleIndex]
        finalBonds = finalNeighbors - finalCenters[refParticleIndex]

        # More succinct notation for doing the calculation, from Bin's original code
        # Converting to numpy matrices makes matrix multiplication happen automatically
        b0 = np.mat(initialBonds)
        b = np.mat(finalBonds)

        # Calculate the two functions used to minimize D2, X and Y (eq. 2.12 and 2.13 respectively)
        X = b0.transpose() * b
        Y = b0.transpose() * b0

        # Calculate the uniform strain tensor that minimizes D2 (eq. 2.14)
        # Note that we don't include the kronecker delta function since it will
        # be cancelled out when plugged into the D2min equation (eq. 2.11).
        # Also not that this is actually the transpose of the strain tensor as
        # it is defined in the paper, since it makes the matrix multiplication easier
        # in the next step
        epsilon = inv(Y) * X

        # Non-affine part, or the terms that are squared and summed over in eq. 2.11
        non_affine = b - b0*epsilon

        # The final value
        d2min = np.sum(np.square(non_affine))

        if normalize:
            d2min /= len(initialNeighbors)

        # Since we have been working with the transpose of the strain tensor,
        # we have to transpose to get the proper one
        return (d2min, np.array(epsilon.transpose()))

    # If we don't have a reference particle, or we are given multiple, we calculate for each of those
    if not isinstance(refParticleIndex, list) and not isinstance(refParticleIndex, np.ndarray):
        refParticleIndex = np.arange(N)

    # Now calculate for all of those possibilities
    d2minArr = np.zeros(len(refParticleIndex))
    epsilonArr = np.zeros([len(refParticleIndex), d, d])
    
    for i in range(len(refParticleIndex)):

        d2min, epsilon = calculateD2Min(initialCenters, finalCenters, refParticleIndex[i],
                                       interactionRadius, interactionNeighbors, normalize)
        d2minArr[i] = d2min
        epsilonArr[i] = epsilon

    return (d2minArr, epsilonArr)

        
def vonMisesStrain(uniformStrainTensor):
    """
    WIP.

    """
    # The number of spatial dimensions
    dimension = np.shape(uniformStrainTensor)[0]

    # Lagrangian strain matrix
    eta = 0.5 * (uniformStrainTensor * uniformStrainTensor.transpose() - np.eye(dimension))

    # von-Mises strain
    eta_m = 1.0/np.double(dimension) * np.trace(eta)
    tmp = eta - eta_m * np.eye(dimension)
    eta_s = np.sqrt(0.5*np.trace(tmp*tmp))

    return eta_s
